import json
import logging
from unittest import mock

import numpy as np
from google.protobuf import any_pb2

from seldon_core.proto import prediction_pb2
from seldon_core.seldon_client import (
    SeldonClient,
    SeldonClientCombine,
    SeldonClientPrediction,
)
from seldon_core.utils import (
    array_to_grpc_datadef,
    json_to_seldon_message,
    seldon_message_to_json,
)

JSON_TEST_DATA = {"test": [0.0, 1.0]}
RAW_DATA_TEST = {"data": {"ndarray": [[0.0, 1.0, 2.0]]}}
CUSTOM_TEST_DATA = any_pb2.Any(value=b"test")


class MockResponse:
    def __init__(self, json_data, status_code, reason="", text=""):
        self.json_data = json_data
        self.status_code = status_code
        self.reason = reason
        self.text = text

    def json(self):
        return self.json_data


def mocked_requests_post_404(url, *args, **kwargs):
    return MockResponse(None, 404, "Not Found")


def mocked_requests_post_success(url, *args, **kwargs):
    data = np.random.rand(1, 1)
    datadef = array_to_grpc_datadef("tensor", data)
    request = prediction_pb2.SeldonMessage(data=datadef)
    json = seldon_message_to_json(request)
    return MockResponse(json, 200, text="{}")


def mocked_requests_post_success_json_data(url, *args, **kwargs):
    request = json_to_seldon_message({"jsonData": JSON_TEST_DATA})
    json = seldon_message_to_json(request)
    return MockResponse(json, 200, text="{}")


@mock.patch("requests.post", side_effect=mocked_requests_post_404)
def test_predict_rest_404(mock_post):
    sc = SeldonClient(deployment_name="404")
    response = sc.predict()
    assert response.success == False
    assert response.msg == "404:Not Found"


@mock.patch("requests.post", side_effect=mocked_requests_post_success)
def test_predict_rest(mock_post):
    sc = SeldonClient(deployment_name="mymodel")
    response = sc.predict(client_return_type="proto")
    logging.info(mock_post.call_args)
    assert response.success == True
    assert response.response.data.tensor.shape == [1, 1]
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success)
def test_predict_rest_with_names(mock_post):
    sc = SeldonClient(deployment_name="mymodel")
    response = sc.predict(names=["a", "b"], client_return_type="proto")
    assert mock_post.call_args[1]["json"]["data"]["names"] == ["a", "b"]
    assert response.success == True
    assert response.response.data.tensor.shape == [1, 1]
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success_json_data)
def test_predict_rest_json_data_ambassador(mock_post):
    sc = SeldonClient(deployment_name="mymodel", gateway="ambassador")
    response = sc.predict(json_data=JSON_TEST_DATA, client_return_type="proto")
    json_response = seldon_message_to_json(response.response)
    assert "jsonData" in mock_post.call_args[1]["json"]
    assert mock_post.call_args[1]["json"]["jsonData"] == JSON_TEST_DATA
    assert response.success is True
    assert json_response["jsonData"] == JSON_TEST_DATA
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success_json_data)
def test_predict_rest_json_data_ambassador_dict_response(mock_post):
    sc = SeldonClient(
        deployment_name="mymodel", gateway="ambassador", client_return_type="dict"
    )
    response = sc.predict(json_data=JSON_TEST_DATA)
    json_response = response.response
    assert "jsonData" in mock_post.call_args[1]["json"]
    assert mock_post.call_args[1]["json"]["jsonData"] == JSON_TEST_DATA
    assert response.success is True
    assert json_response["jsonData"] == JSON_TEST_DATA
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success_json_data)
def test_predict_rest_json_data_seldon(mock_post):
    sc = SeldonClient(
        deployment_name="mymodel", gateway="seldon", client_return_type="proto"
    )
    response = sc.predict(json_data=JSON_TEST_DATA)
    json_response = seldon_message_to_json(response.response)
    assert "jsonData" in mock_post.call_args[1]["json"]
    assert mock_post.call_args[1]["json"]["jsonData"] == JSON_TEST_DATA
    assert response.success is True
    assert json_response["jsonData"] == JSON_TEST_DATA
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success)
def test_predict_rest_raw_data_seldon_proto(mock_post):
    sc = SeldonClient(
        deployment_name="mymodel", gateway="seldon", client_return_type="proto"
    )
    response = sc.predict(raw_data=RAW_DATA_TEST)
    json_response = seldon_message_to_json(response.response)
    assert mock_post.call_args[1]["json"] == RAW_DATA_TEST
    assert response.success is True
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success)
def test_predict_rest_raw_data_gateway_dict(mock_post):
    sc = SeldonClient(
        deployment_name="mymodel", gateway="istio", client_return_type="dict"
    )
    response = sc.predict(raw_data=RAW_DATA_TEST)
    json_response = response.response
    assert mock_post.call_args[1]["json"] == RAW_DATA_TEST
    assert response.success is True
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success_json_data)
def test_predict_rest_json_data_seldon_return_type(mock_post):
    sc = SeldonClient(
        deployment_name="mymodel", gateway="seldon", client_return_type="dict"
    )
    response = sc.predict(json_data=JSON_TEST_DATA)
    json_response = response.response
    assert "jsonData" in mock_post.call_args[1]["json"]
    assert mock_post.call_args[1]["json"]["jsonData"] == JSON_TEST_DATA
    assert response.success is True
    assert json_response["jsonData"] == JSON_TEST_DATA
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success_json_data)
def test_explain_rest_json_data_ambassador(mock_post):
    sc = SeldonClient(
        deployment_name="mymodel", gateway="ambassador", client_return_type="dict"
    )
    response = sc.explain(json_data=JSON_TEST_DATA, predictor="default")
    json_response = response.response
    # Currently this doesn't need to convert to JSON due to #1083
    # i.e. json_response = seldon_message_to_json(response.response)
    assert "jsonData" in mock_post.call_args[1]["json"]
    assert mock_post.call_args[1]["json"]["jsonData"] == JSON_TEST_DATA
    assert json_response["jsonData"] == JSON_TEST_DATA
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success_json_data)
def test_explain_rest_json_data_ambassador_dict_response(mock_post):
    sc = SeldonClient(
        deployment_name="mymodel", gateway="ambassador", client_return_type="dict"
    )
    response = sc.explain(json_data=JSON_TEST_DATA, predictor="default")
    json_response = response.response
    # Currently this doesn't need to convert to JSON due to #1083
    # i.e. json_response = seldon_message_to_json(response.response)
    assert "jsonData" in mock_post.call_args[1]["json"]
    assert mock_post.call_args[1]["json"]["jsonData"] == JSON_TEST_DATA
    assert json_response["jsonData"] == JSON_TEST_DATA
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success)
def test_predict_rest_with_meta(mock_post):
    sc = SeldonClient(deployment_name="mymodel")
    meta = {"key": "value"}
    response = sc.predict(names=["a", "b"], meta=meta, client_return_type="proto")
    assert mock_post.call_args[1]["json"]["data"]["names"] == ["a", "b"]
    assert mock_post.call_args[1]["json"]["meta"]["tags"] == meta
    assert response.success == True
    assert response.response.data.tensor.shape == [1, 1]
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success)
def test_predict_rest_with_ambassador_prefix(mock_post):
    sc = SeldonClient(deployment_name="mymodel")
    response = sc.predict(
        gateway="ambassador",
        transport="rest",
        gateway_prefix="/mycompany/ml",
        client_return_type="proto",
    )
    assert mock_post.call_args[0][0].index("/mycompany/ml") > 0
    assert response.success == True
    assert response.response.data.tensor.shape == [1, 1]
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success)
def test_predict_rest_with_ambassador_prefix_dict_response(mock_post):
    sc = SeldonClient(deployment_name="mymodel", client_return_type="dict")
    response = sc.predict(
        gateway="ambassador", transport="rest", gateway_prefix="/mycompany/ml"
    )
    assert mock_post.call_args[0][0].index("/mycompany/ml") > 0
    assert response.success == True
    assert response.response["data"]["tensor"]["shape"] == [1, 1]
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success)
def test_predict_microservice_rest(mock_post):
    sc = SeldonClient(deployment_name="mymodel")
    response = sc.microservice(method="predict")
    logging.info(response)
    assert response.success == True
    assert response.response.data.tensor.shape == [1, 1]
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success_json_data)
def test_predict_microservice_rest_json_data(mock_post):
    sc = SeldonClient(deployment_name="mymodel")
    response = sc.microservice(method="predict", json_data=JSON_TEST_DATA)
    json_response = seldon_message_to_json(response.response)
    assert "jsonData" in mock_post.call_args[1]["data"]["json"]
    assert response.success is True
    assert mock_post.call_args[1]["data"]["json"] == json.dumps(
        {"jsonData": JSON_TEST_DATA}
    )
    assert json_response["jsonData"] == JSON_TEST_DATA
    assert mock_post.call_count == 1


@mock.patch("requests.post", side_effect=mocked_requests_post_success)
def test_feedback_microservice_rest(mock_post):
    sc = SeldonClient(deployment_name="mymodel")
    response = sc.microservice_feedback(
        prediction_request=prediction_pb2.SeldonMessage(),
        prediction_response=prediction_pb2.SeldonMessage(),
        reward=1.0,
    )
    assert response.success == True
    assert response.response.data.tensor.shape == [1, 1]
    assert mock_post.call_count == 1


class MyStub:
    def __init__(self, channel):
        self.channel = channel

    def Predict(self, **kwargs):
        return prediction_pb2.SeldonMessage(strData="predict")

    def TransformInput(selfself, **kwargs):
        return prediction_pb2.SeldonMessage(strData="transform-input")

    def TransformOutput(selfself, **kwargs):
        return prediction_pb2.SeldonMessage(strData="transform-output")

    def Route(selfself, **kwargs):
        return prediction_pb2.SeldonMessage(strData="route")


def mock_grpc_stub_predict(channel):
    return MyStub()


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.SeldonStub", new=MyStub)
def test_predict_grpc_ambassador():
    sc = SeldonClient(deployment_name="mymodel", transport="grpc", gateway="ambassador")
    response = sc.predict(client_return_type="proto")
    assert response.response.strData == "predict"


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.SeldonStub", new=MyStub)
def test_predict_grpc_ambassador_with_meta():
    sc = SeldonClient(deployment_name="mymodel", transport="grpc", gateway="ambassador")
    response = sc.predict(meta={"key": "value"}, client_return_type="proto")
    assert response.response.strData == "predict"


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.SeldonStub", new=MyStub)
def test_grpc_predict_json_data_ambassador():
    sc = SeldonClient(deployment_name="mymodel", transport="grpc", gateway="ambassador")
    response = sc.predict(json_data=JSON_TEST_DATA, client_return_type="proto")
    assert response.response.strData == "predict"


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.SeldonStub", new=MyStub)
def test_grpc_predict_custom_data_ambassador():
    sc = SeldonClient(deployment_name="mymodel", transport="grpc", gateway="ambassador")
    response = sc.predict(custom_data=CUSTOM_TEST_DATA, client_return_type="proto")
    assert response.response.strData == "predict"


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.SeldonStub", new=MyStub)
def test_predict_grpc_seldon():
    sc = SeldonClient(deployment_name="mymodel", transport="grpc", gateway="seldon")
    response = sc.predict(client_return_type="proto")
    assert response.response.strData == "predict"


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.SeldonStub", new=MyStub)
def test_predict_grpc_proto_raw_data_seldon():
    sc = SeldonClient(deployment_name="mymodel", transport="grpc", gateway="seldon")
    proto_raw_data = json_to_seldon_message(RAW_DATA_TEST)
    response = sc.predict(raw_data=proto_raw_data, client_return_type="proto")
    request = seldon_message_to_json(response.request)
    assert request == RAW_DATA_TEST


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.SeldonStub", new=MyStub)
def test_predict_grpc_raw_data_gateway():
    sc = SeldonClient(deployment_name="mymodel", transport="grpc", gateway="istio")
    response = sc.predict(raw_data=RAW_DATA_TEST, client_return_type="proto")
    request = seldon_message_to_json(response.request)
    assert request == RAW_DATA_TEST


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.SeldonStub", new=MyStub)
def test_grpc_predict_json_data_seldon():
    sc = SeldonClient(
        deployment_name="mymodel",
        transport="grpc",
        gateway="seldon",
        client_return_type="proto",
    )
    response = sc.predict(json_data=JSON_TEST_DATA)
    assert response.response.strData == "predict"


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.SeldonStub", new=MyStub)
def test_grpc_predict_custom_data_seldon():
    sc = SeldonClient(
        deployment_name="mymodel",
        transport="grpc",
        gateway="seldon",
        client_return_type="proto",
    )
    response = sc.predict(custom_data=CUSTOM_TEST_DATA)
    assert response.response.strData == "predict"


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.ModelStub", new=MyStub)
def test_predict_grpc_microservice_predict():
    sc = SeldonClient(transport="grpc")
    response = sc.microservice(method="predict")
    assert response.response.strData == "predict"


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.GenericStub", new=MyStub)
def test_predict_grpc_microservice_transform_input():
    sc = SeldonClient(transport="grpc")
    response = sc.microservice(method="transform-input")
    assert response.response.strData == "transform-input"


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.GenericStub", new=MyStub)
def test_predict_grpc_microservice_transform_output():
    sc = SeldonClient(transport="grpc")
    response = sc.microservice(method="transform-output")
    assert response.response.strData == "transform-output"


@mock.patch("seldon_core.seldon_client.prediction_pb2_grpc.GenericStub", new=MyStub)
def test_predict_grpc_microservice_transform_route():
    sc = SeldonClient(transport="grpc")
    response = sc.microservice(method="route")
    assert response.response.strData == "route"


#
# Wiring Tests
#


@mock.patch(
    "seldon_core.seldon_client.microservice_api_rest_seldon_message",
    return_value=SeldonClientPrediction(None, None),
)
def test_wiring_microservice_api_rest_seldon_message(mock_handler):
    sc = SeldonClient()
    response = sc.microservice(transport="rest", method="predict")
    assert mock_handler.call_count == 1


@mock.patch(
    "seldon_core.seldon_client.microservice_api_rest_aggregate",
    return_value=SeldonClientCombine(None, None),
)
def test_wiring_microservice_api_rest_aggregate(mock_handler):
    sc = SeldonClient()
    response = sc.microservice(transport="rest", method="aggregate")
    assert mock_handler.call_count == 1


@mock.patch(
    "seldon_core.seldon_client.microservice_api_rest_feedback",
    return_value=SeldonClientCombine(None, None),
)
def test_wiring_microservice_api_rest_feedback(mock_handler):
    sc = SeldonClient()
    response = sc.microservice_feedback(
        prediction_pb2.SeldonMessage(),
        prediction_pb2.SeldonMessage(),
        1.0,
        transport="rest",
    )
    assert mock_handler.call_count == 1


@mock.patch(
    "seldon_core.seldon_client.microservice_api_grpc_seldon_message",
    return_value=SeldonClientPrediction(None, None),
)
def test_wiring_microservice_api_grpc_seldon_message(mock_handler):
    sc = SeldonClient()
    response = sc.microservice(transport="grpc", method="predict")
    assert mock_handler.call_count == 1


@mock.patch(
    "seldon_core.seldon_client.microservice_api_grpc_aggregate",
    return_value=SeldonClientCombine(None, None),
)
def test_wiring_microservice_api_grpc_aggregate(mock_handler):
    sc = SeldonClient()
    response = sc.microservice(transport="grpc", method="aggregate")
    assert mock_handler.call_count == 1


@mock.patch(
    "seldon_core.seldon_client.microservice_api_grpc_feedback",
    return_value=SeldonClientCombine(None, None),
)
def test_wiring_microservice_api_grpc_feedback(mock_handler):
    sc = SeldonClient()
    response = sc.microservice_feedback(
        prediction_pb2.SeldonMessage(),
        prediction_pb2.SeldonMessage(),
        1.0,
        transport="grpc",
    )
    assert mock_handler.call_count == 1


@mock.patch(
    "seldon_core.seldon_client.rest_predict_gateway",
    return_value=SeldonClientPrediction(None, None),
)
def test_wiring_rest_predict_ambassador(mock_rest_predict_ambassador):
    sc = SeldonClient(deployment_name="mymodel")
    response = sc.predict(gateway="ambassador", transport="rest")
    assert mock_rest_predict_ambassador.call_count == 1


@mock.patch(
    "seldon_core.seldon_client.grpc_predict_gateway",
    return_value=SeldonClientPrediction(None, None),
)
def test_wiring_grpc_predict_ambassador(mock_grpc_predict_ambassador):
    sc = SeldonClient(deployment_name="mymodel")
    response = sc.predict(gateway="ambassador", transport="grpc")
    assert mock_grpc_predict_ambassador.call_count == 1



import argparse
import contextlib
import importlib
import json
import logging
import os
import socket
import sys
import time
from distutils.util import strtobool
from functools import partial
from typing import Any, Callable, Dict, List, Tuple

from seldon_core import __version__
from seldon_core import wrapper as seldon_microservice
from seldon_core.flask_utils import (
    ANNOTATION_REST_TIMEOUT,
    ANNOTATIONS_FILE,
    DEFAULT_ANNOTATION_REST_TIMEOUT,
    SeldonMicroserviceException,
)
from seldon_core.gunicorn_utils import (
    StandaloneApplication,
    UserModelApplication,
    accesslog,
    post_worker_init,
    threads,
    worker_exit,
)
from seldon_core.metrics import SeldonMetrics
from seldon_core.utils import getenv_as_bool, setup_tracing

# This is related to how multiprocessing is implemeneted on MacOS
# See https://github.com/SeldonIO/seldon-core/issues/3410 for discussion.
USE_MULTIPROCESS_ENV_NAME = "USE_MULTIPROCESS_PACKAGE"
USE_MULTIPROCESS = getenv_as_bool(USE_MULTIPROCESS_ENV_NAME, default=False)
if USE_MULTIPROCESS:
    import multiprocess as mp
else:
    import multiprocessing as mp


logger = logging.getLogger(__name__)

PARAMETERS_ENV_NAME = "PREDICTIVE_UNIT_PARAMETERS"
HTTP_SERVICE_PORT_ENV_NAME = "PREDICTIVE_UNIT_HTTP_SERVICE_PORT"
GRPC_SERVICE_PORT_ENV_NAME = "PREDICTIVE_UNIT_GRPC_SERVICE_PORT"
METRICS_SERVICE_PORT_ENV_NAME = "PREDICTIVE_UNIT_METRICS_SERVICE_PORT"

FILTER_METRICS_ACCESS_LOGS_ENV_NAME = "FILTER_METRICS_ACCESS_LOGS"

LOG_LEVEL_ENV = "SELDON_LOG_LEVEL"
DEFAULT_LOG_LEVEL = "INFO"

DEFAULT_GRPC_PORT = 5000
DEFAULT_HTTP_PORT = 9000
DEFAULT_METRICS_PORT = 6000

DEBUG_ENV = "SELDON_DEBUG"
GUNICORN_ACCESS_LOG_ENV = "GUNICORN_ACCESS_LOG"


def start_servers(
    target1: Callable, target2: Callable, target3: Callable, metrics_target: Callable
) -> None:
    """
    Start servers

    Parameters
    ----------
    target1
       Main flask process
    target2
       Auxiliary flask process

    """
    if USE_MULTIPROCESS:
        logger.info("Using alternative multiprocessing library")
    else:
        logger.info("Using standard multiprocessing library")

    p2 = None
    if target2:
        p2 = mp.Process(target=target2, daemon=False)
        p2.start()

    p3 = None
    if target3:
        p3 = mp.Process(target=target3, daemon=True)
        p3.start()

    p4 = None
    if metrics_target:
        p4 = mp.Process(target=metrics_target, daemon=True)
        p4.start()

    target1()

    if p2:
        p2.join()

    if p3:
        p3.join()

    if p4:
        p4.join()


def parse_parameters(parameters: Dict) -> Dict:
    """
    Parse the user object parameters

    Parameters
    ----------
    parameters

    Returns
    -------

    """
    type_dict = {
        "INT": int,
        "FLOAT": float,
        "DOUBLE": float,
        "STRING": str,
        "BOOL": bool,
    }
    parsed_parameters = {}
    for param in parameters:
        name = param.get("name")
        value = param.get("value")
        type_ = param.get("type")
        if type_ == "BOOL":
            parsed_parameters[name] = bool(strtobool(value))
        else:
            try:
                parsed_parameters[name] = type_dict[type_](value)
            except ValueError:
                raise SeldonMicroserviceException(
                    "Bad model parameter: "
                    + name
                    + " with value "
                    + value
                    + " can't be parsed as a "
                    + type_,
                    reason="MICROSERVICE_BAD_PARAMETER",
                )
            except KeyError:
                raise SeldonMicroserviceException(
                    "Bad model parameter type: "
                    + type_
                    + " valid are INT, FLOAT, DOUBLE, STRING, BOOL",
                    reason="MICROSERVICE_BAD_PARAMETER",
                )
    return parsed_parameters


def load_annotations() -> Dict[str, str]:
    """
    Attempt to load annotations

    Returns
    -------

    """
    annotations = {}
    try:
        if os.path.isfile(ANNOTATIONS_FILE):
            with open(ANNOTATIONS_FILE, "r") as ins:
                for line in ins:
                    line = line.rstrip()
                    parts = list(map(str.strip, line.split("=", 1)))
                    if len(parts) == 2:
                        key = parts[0]
                        value = parts[1][1:-1]  # strip quotes at start and end
                        logger.info("Found annotation %s:%s ", key, value)
                        annotations[key] = value
                    else:
                        logger.info("Bad annotation [%s]", line)
    except:
        logger.error("Failed to open annotations file %s", ANNOTATIONS_FILE)
    return annotations


class MetricsEndpointFilter(logging.Filter):
    def filter(self, record):
        return seldon_microservice.METRICS_ENDPOINT not in record.getMessage()


def setup_logger(log_level: str, debug_mode: bool) -> logging.Logger:
    # set up log level
    log_level_raw = os.environ.get(LOG_LEVEL_ENV, log_level.upper())
    log_level_num = getattr(logging, log_level_raw, None)
    if not isinstance(log_level_num, int):
        raise ValueError("Invalid log level: %s", log_level)

    logger.setLevel(log_level_num)

    # Set right level on access logs
    flask_logger = logging.getLogger("werkzeug")
    flask_logger.setLevel(log_level_num)

    if getenv_as_bool(FILTER_METRICS_ACCESS_LOGS_ENV_NAME, default=not debug_mode):
        flask_logger.addFilter(MetricsEndpointFilter())
        gunicorn_logger = logging.getLogger("gunicorn.access")
        gunicorn_logger.addFilter(MetricsEndpointFilter())

    logger.debug("Log level set to %s:%s", log_level, log_level_num)

    # set log level for the imported microservice type
    seldon_microservice.logger.setLevel(log_level_num)
    logging.getLogger().setLevel(log_level_num)
    for handler in logger.handlers:
        handler.setLevel(log_level_num)

    return logger


def parse_args() -> Tuple[argparse.Namespace, List[str]]:
    parser = argparse.ArgumentParser()
    parser.add_argument("interface_name", type=str, help="Name of the user interface.")

    parser.add_argument(
        "--service-type",
        type=str,
        choices=["MODEL", "ROUTER", "TRANSFORMER", "COMBINER", "OUTLIER_DETECTOR"],
        default="MODEL",
    )
    parser.add_argument(
        "--persistence",
        nargs="?",
        default=0,
        const=1,
        type=int,
        help="deprecated argument ",
    )
    parser.add_argument(
        "--parameters", type=str, default=os.environ.get(PARAMETERS_ENV_NAME, "[]")
    )
    parser.add_argument(
        "--log-level",
        type=str,
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        default=DEFAULT_LOG_LEVEL,
        help="Log level of the inference server.",
    )
    parser.add_argument(
        "--debug",
        nargs="?",
        type=bool,
        default=getenv_as_bool(DEBUG_ENV, default=False),
        const=True,
        help="Enable debug mode.",
    )
    parser.add_argument(
        "--tracing",
        nargs="?",
        default=int(os.environ.get("TRACING", "0")),
        const=1,
        type=int,
    )

    # gunicorn settings, defaults are from
    # http://docs.gunicorn.org/en/stable/settings.html
    parser.add_argument(
        "--workers",
        type=int,
        default=int(os.environ.get("GUNICORN_WORKERS", "1")),
        help="Number of Gunicorn workers for handling requests.",
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=int(os.environ.get("GUNICORN_THREADS", "1")),
        help="Number of threads to run per Gunicorn worker.",
    )
    parser.add_argument(
        "--max-requests",
        type=int,
        default=int(os.environ.get("GUNICORN_MAX_REQUESTS", "0")),
        help="Maximum number of requests gunicorn worker will process before restarting.",
    )
    parser.add_argument(
        "--max-requests-jitter",
        type=int,
        default=int(os.environ.get("GUNICORN_MAX_REQUESTS_JITTER", "0")),
        help="Maximum random jitter to add to max-requests.",
    )
    parser.add_argument(
        "--keepalive",
        type=int,
        default=int(os.environ.get("GUNICORN_KEEPALIVE", "2")),
        help="The number of seconds to wait for requests on a Keep-Alive connection.",
    )

    parser.add_argument(
        "--single-threaded",
        type=int,
        default=int(os.environ.get("FLASK_SINGLE_THREADED", "0")),
        help="Force the Flask app to run single-threaded. Also applies to Gunicorn.",
    )

    parser.add_argument(
        "--http-port",
        type=int,
        default=int(os.environ.get(HTTP_SERVICE_PORT_ENV_NAME, DEFAULT_HTTP_PORT)),
        help="Set http port of seldon service",
    )

    parser.add_argument(
        "--grpc-port",
        type=int,
        default=int(os.environ.get(GRPC_SERVICE_PORT_ENV_NAME, DEFAULT_GRPC_PORT)),
        help="Set grpc port of seldon service",
    )

    parser.add_argument(
        "--metrics-port",
        type=int,
        default=int(
            os.environ.get(METRICS_SERVICE_PORT_ENV_NAME, DEFAULT_METRICS_PORT)
        ),
        help="Set metrics port of seldon service",
    )

    parser.add_argument(
        "--pidfile", type=str, default=None, help="A file path to use for the PID file"
    )

    parser.add_argument(
        "--access-log",
        nargs="?",
        type=bool,
        default=getenv_as_bool(GUNICORN_ACCESS_LOG_ENV, default=False),
        const=True,
        help="Enable gunicorn access log.",
    )

    parser.add_argument(
        "--grpc-threads",
        type=int,
        default=os.environ.get("GRPC_THREADS", default="1"),
        help="Number of GRPC threads per worker.",
    )

    parser.add_argument(
        "--grpc-workers",
        type=int,
        default=os.environ.get("GRPC_WORKERS", default="1"),
        help="Number of GPRC workers.",
    )

    return parser.parse_known_args()


def _make_rest_server_debug(
    user_object: Any,
    seldon_metrics: SeldonMetrics,
    args: argparse.Namespace,
    jaeger_extra_tags: List[str],
) -> Callable[[], None]:
    """Makes a function that creates a REST debugging server.
    Args:
        user_object: an instance of user-defined class, inherited from user_model.SeldonComponent.
        seldon_metrics: a SeldonMetrics instance.
        args: parsed args from commandline.
        jaeger_extra_tags:
    """

    def server():
        app = seldon_microservice.get_rest_microservice(user_object, seldon_metrics)
        try:
            user_object.load()
        except (NotImplementedError, AttributeError):
            pass
        if args.tracing:
            logger.info("Tracing branch is active")
            from flask_opentracing import FlaskTracing

            tracer = setup_tracing(args.interface_name)

            logger.info("Set JAEGER_EXTRA_TAGS %s", jaeger_extra_tags)
            FlaskTracing(tracer, True, app, jaeger_extra_tags)

        # Timeout not supported in flask development server
        app.run(
            host="0.0.0.0",
            port=args.http_port,
            threaded=False if args.single_threaded else True,
        )

    return server


def _make_rest_server_prod(
    user_object: Any,
    seldon_metrics: SeldonMetrics,
    args: argparse.Namespace,
    jaeger_extra_tags: List[str],
    annotations: Dict[str, str],
) -> Callable[[], None]:
    """Makes a function that creates a REST production server.
    Args:
        user_object: an instance of user-defined class, inherited from user_model.SeldonComponent.
        seldon_metrics: a SeldonMetrics instance.
        args: parsed args from commandline.
        jaeger_extra_tags:
        annotations:
    """

    def server() -> None:
        rest_timeout = DEFAULT_ANNOTATION_REST_TIMEOUT
        if ANNOTATION_REST_TIMEOUT in annotations:
            # Gunicorn timeout is in seconds so convert as annotation is in miliseconds
            rest_timeout = int(annotations[ANNOTATION_REST_TIMEOUT]) / 1000
            # Converting timeout from float to int and set to 1 if is 0
            rest_timeout = int(rest_timeout) or 1

        options = {
            "bind": "%s:%s" % ("0.0.0.0", args.http_port),
            "accesslog": accesslog(args.access_log),
            "loglevel": args.log_level.lower(),
            "timeout": rest_timeout,
            "threads": threads(args.threads, args.single_threaded),
            "workers": args.workers,
            "max_requests": args.max_requests,
            "max_requests_jitter": args.max_requests_jitter,
            "post_worker_init": post_worker_init,
            "worker_exit": partial(worker_exit, seldon_metrics=seldon_metrics),
            "keepalive": args.keepalive,
        }
        logger.info(f"Gunicorn Config:  {options}")

        if args.pidfile is not None:
            options["pidfile"] = args.pidfile
        app = seldon_microservice.get_rest_microservice(user_object, seldon_metrics)

        UserModelApplication(
            app,
            user_object,
            args.tracing,
            jaeger_extra_tags,
            args.interface_name,
            options=options,
        ).run()

    return server


def _wait_forever(server):
    try:
        while True:
            time.sleep(60 * 60)
    except KeyboardInterrupt:
        server.stop(None)


def _run_grpc_server(
    user_object: Any,
    seldon_metrics: SeldonMetrics,
    args: argparse.Namespace,
    annotations: Dict[str, str],
    bind_address: str,
):
    """Start a server in a subprocess."""
    logger.info(f"Starting new GRPC server with {args.grpc_threads} threads.")

    if args.tracing:
        from grpc_opentracing import open_tracing_server_interceptor

        logger.info("Adding tracer")
        tracer = setup_tracing(args.interface_name)
        interceptor = open_tracing_server_interceptor(tracer)
    else:
        interceptor = None

    server = seldon_microservice.get_grpc_server(
        user_object,
        seldon_metrics,
        annotations=annotations,
        trace_interceptor=interceptor,
        num_threads=args.grpc_threads,
    )

    try:
        user_object.load()
    except (NotImplementedError, AttributeError):
        pass

    server.add_insecure_port(bind_address)
    server.start()
    _wait_forever(server)


def _make_grpc_server(
    user_object: Any,
    seldon_metrics: SeldonMetrics,
    args: argparse.Namespace,
    annotations: Dict[str, str],
) -> Callable[[], None]:
    def server() -> None:
        with _reserve_grpc_port(args.grpc_port) as bind_port:
            bind_address = "0.0.0.0:{}".format(bind_port)
            logger.info(
                "GRPC Server Binding to %s with %d processes.",
                bind_address,
                args.grpc_workers,
            )
            sys.stdout.flush()
            workers = []
            for _ in range(args.grpc_workers):
                # NOTE: It is imperative that the worker subprocesses be forked before
                # any gRPC servers start up. See
                # https://github.com/grpc/grpc/issues/16001 for more details.
                worker = mp.Process(
                    target=_run_grpc_server,
                    args=(
                        user_object,
                        seldon_metrics,
                        args,
                        annotations,
                        bind_address,
                    ),
                )
                worker.start()
                workers.append(worker)
            for worker in workers:
                worker.join()

    return server


def _make_rest_metrics_server(
    seldon_metrics: SeldonMetrics,
    args: argparse.Namespace,
) -> Callable[[], None]:
    def server() -> None:
        app = seldon_microservice.get_metrics_microservice(seldon_metrics)
        if args.debug:
            app.run(host="0.0.0.0", port=args.metrics_port)
        else:
            options = {
                "bind": "%s:%s" % ("0.0.0.0", args.metrics_port),
                "accesslog": accesslog(args.access_log),
                "loglevel": args.log_level.lower(),
                "timeout": 5000,
                "max_requests": args.max_requests,
                "max_requests_jitter": args.max_requests_jitter,
                "post_worker_init": post_worker_init,
                "keepalive": args.keepalive,
            }
            if args.pidfile is not None:
                options["pidfile"] = args.pidfile
            StandaloneApplication(app, options=options).run()

    return server


@contextlib.contextmanager
def _reserve_grpc_port(grpc_port: int):
    """Find and reserve a port for all subprocesses to use."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    if sock.getsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT) != 1:
        raise RuntimeError("Failed to set SO_REUSEPORT.")
    sock.bind(("", grpc_port))
    try:
        yield sock.getsockname()[1]
    finally:
        sock.close()


def main():
    LOG_FORMAT = (
        "%(asctime)s - %(name)s:%(funcName)s:%(lineno)s - %(levelname)s:  %(message)s"
    )
    logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
    logger.info("Starting microservice.py:main")
    logger.info(f"Seldon Core version: {__version__}")

    sys.path.append(os.getcwd())

    args, remaining = parse_args()

    if len(remaining) > 0:
        logger.error(
            f"Unknown args {remaining}. Note since 1.5.0 this CLI does not take API type (REST, GRPC)"
        )
        sys.exit(-1)

    parameters = parse_parameters(json.loads(args.parameters))

    setup_logger(args.log_level, args.debug)

    # set flask trace jaeger extra tags
    jaeger_extra_tags = list(
        filter(
            lambda x: (x != ""),
            [tag.strip() for tag in os.environ.get("JAEGER_EXTRA_TAGS", "").split(",")],
        )
    )
    logger.info("Parse JAEGER_EXTRA_TAGS %s", jaeger_extra_tags)

    annotations = load_annotations()
    logger.info("Annotations: %s", annotations)

    parts = args.interface_name.rsplit(".", 1)
    if len(parts) == 1:
        logger.info("Importing %s", args.interface_name)
        interface_file = importlib.import_module(args.interface_name)
        user_class = getattr(interface_file, args.interface_name)
    else:
        logger.info("Importing submodule %s", parts)
        interface_file = importlib.import_module(parts[0])
        user_class = getattr(interface_file, parts[1])

    if args.persistence:
        logger.error(f"Persistence: ignored, persistence is deprecated")
    user_object = user_class(**parameters)

    http_port = args.http_port

    seldon_metrics = SeldonMetrics(worker_id_func=os.getpid)
    # TODO why 2 ways to create metrics server
    # seldon_metrics = SeldonMetrics(
    #    worker_id_func=lambda: threading.current_thread().name
    # )
    if args.debug:
        # Start Flask debug server
        logger.info(
            "REST microservice running on port %i single-threaded=%s",
            http_port,
            args.single_threaded,
        )
        server_rest_func = _make_rest_server_debug(
            user_object, seldon_metrics, args, jaeger_extra_tags=jaeger_extra_tags
        )
    else:
        # Start production server
        logger.info("REST gunicorn microservice running on port %i", http_port)
        server_rest_func = _make_rest_server_prod(
            user_object,
            seldon_metrics,
            args,
            jaeger_extra_tags=jaeger_extra_tags,
            annotations=annotations,
        )

    server_grpc_func = None
    if args.grpc_workers > 0:
        server_grpc_func = _make_grpc_server(
            user_object, seldon_metrics, args, annotations=annotations
        )

    logger.info("REST metrics microservice running on port %i", args.metrics_port)
    server_metrics_func = _make_rest_metrics_server(seldon_metrics, args)

    if hasattr(user_object, "custom_service") and callable(
        getattr(user_object, "custom_service")
    ):
        server_custom_func = user_object.custom_service
    else:
        server_custom_func = None

    logger.info("Starting servers")
    start_servers(
        server_rest_func,
        server_grpc_func,
        server_custom_func,
        server_metrics_func,
    )


if __name__ == "__main__":
    main()



import inspect
import json
import logging
import os
from typing import Dict, Iterable, List, Tuple, Union

import numpy as np

from seldon_core.flask_utils import SeldonMicroserviceException
from seldon_core.metrics import SeldonMetrics, validate_metrics
from seldon_core.proto import prediction_pb2

logger = logging.getLogger(__name__)


INCLUDE_METRICS_IN_CLIENT_RESPONSE = (
    os.environ.get("INCLUDE_METRICS_IN_CLIENT_RESPONSE", "true").lower() == "true"
)


class SeldonNotImplementedError(SeldonMicroserviceException):
    status_code = 400

    def __init__(self, message):
        SeldonMicroserviceException.__init__(self, message)


class SeldonResponse:
    """SeldonResponse

    Simple class to store data returned by SeldonComponent methods together
    with relevant runtime metrics and runtime request tags.
    """

    def __init__(
        self,
        data: Union[np.ndarray, List, str, bytes],
        tags: Dict = None,
        metrics: List[Dict] = None,
    ):
        self.data = data
        self.tags = tags if tags is not None else {}
        self.metrics = metrics if metrics is not None else []

    @classmethod
    def create(
        cls, data: Union[np.ndarray, List, Dict, str, bytes, "SeldonResponse"]
    ) -> "SeldonResponse":
        if isinstance(data, cls):
            return data
        else:
            return cls(data=data)


class SeldonComponent:
    def __init__(self, **kwargs):
        pass

    def tags(self) -> Dict:
        raise SeldonNotImplementedError("tags is not implemented")

    def class_names(self) -> Iterable[str]:
        raise SeldonNotImplementedError("class_names is not implemented")

    def load(self):
        pass

    def predict(
        self, X: np.ndarray, names: Iterable[str], meta: Dict = None
    ) -> Union[np.ndarray, List, Dict, str, bytes, SeldonResponse]:
        raise SeldonNotImplementedError("predict is not implemented")

    def predict_raw(
        self, msg: prediction_pb2.SeldonMessage
    ) -> Union[prediction_pb2.SeldonMessage, Dict]:
        raise SeldonNotImplementedError("predict_raw is not implemented")

    def send_feedback_raw(
        self, feedback: prediction_pb2.Feedback
    ) -> Union[prediction_pb2.SeldonMessage, Dict]:
        raise SeldonNotImplementedError("send_feedback_raw is not implemented")

    def transform_input(
        self, X: np.ndarray, names: Iterable[str], meta: Dict = None
    ) -> Union[np.ndarray, List, Dict, str, bytes, SeldonResponse]:
        raise SeldonNotImplementedError("transform_input is not implemented")

    def transform_input_raw(
        self, msg: prediction_pb2.SeldonMessage
    ) -> Union[prediction_pb2.SeldonMessage, Dict]:
        raise SeldonNotImplementedError("transform_input_raw is not implemented")

    def transform_output(
        self, X: np.ndarray, names: Iterable[str], meta: Dict = None
    ) -> Union[np.ndarray, List, Dict, str, bytes, SeldonResponse]:
        raise SeldonNotImplementedError("transform_output is not implemented")

    def transform_output_raw(
        self, msg: prediction_pb2.SeldonMessage
    ) -> Union[prediction_pb2.SeldonMessage, Dict]:
        raise SeldonNotImplementedError("transform_output_raw is not implemented")

    def metrics(self) -> List[Dict]:
        raise SeldonNotImplementedError("metrics is not implemented")

    def feature_names(self) -> Iterable[str]:
        raise SeldonNotImplementedError("feature_names is not implemented")

    def send_feedback(
        self,
        features: Union[np.ndarray, str, bytes],
        feature_names: Iterable[str],
        reward: float,
        truth: Union[np.ndarray, str, bytes],
        routing: Union[int, None],
    ) -> Union[np.ndarray, List, Dict, str, bytes, None, SeldonResponse]:
        raise SeldonNotImplementedError("send_feedback is not implemented")

    def route(
        self, features: Union[np.ndarray, str, bytes], feature_names: Iterable[str]
    ) -> Union[int, SeldonResponse]:
        raise SeldonNotImplementedError("route is not implemented")

    def route_raw(
        self, msg: prediction_pb2.SeldonMessage
    ) -> Union[prediction_pb2.SeldonMessage, Dict]:
        raise SeldonNotImplementedError("route_raw is not implemented")

    def aggregate(
        self,
        features_list: List[Union[np.ndarray, str, bytes]],
        feature_names_list: List,
    ) -> Union[np.ndarray, List, Dict, str, bytes, SeldonResponse]:
        raise SeldonNotImplementedError("aggregate is not implemented")

    def aggregate_raw(
        self, msgs: prediction_pb2.SeldonMessageList
    ) -> Union[prediction_pb2.SeldonMessage, Dict]:
        raise SeldonNotImplementedError("aggregate_raw is not implemented")

    def health_status(self) -> Union[np.ndarray, List, str, bytes]:
        raise SeldonNotImplementedError("health is not implemented")

    def health_status_raw(self) -> prediction_pb2.SeldonMessage:
        raise SeldonNotImplementedError("health_raw is not implemented")

    def metadata(self) -> Dict:
        raise SeldonNotImplementedError("metadata is not implemented")

    def init_metadata(self) -> Dict:
        raise SeldonNotImplementedError("init_metadata is not implemented")


def client_custom_tags(user_model: SeldonComponent) -> Dict:
    """
    Get tags from user model

    Parameters
    ----------
    user_model

    Returns
    -------
       Dictionary of key value pairs

    """
    if hasattr(user_model, "tags"):
        try:
            return user_model.tags()
        except SeldonNotImplementedError:
            pass
    logger.debug("custom_tags is not implemented")
    return {}


def client_class_names(
    user_model: SeldonComponent, predictions: np.ndarray
) -> Iterable[str]:
    """
    Get class names from user model

    Parameters
    ----------
    user_model
       User defined class instance
    predictions
       Prediction results
    Returns
    -------
       Class names
    """
    if len(predictions.shape) > 1:
        if hasattr(user_model, "class_names"):
            if inspect.ismethod(getattr(user_model, "class_names")):
                try:
                    return user_model.class_names()
                except SeldonNotImplementedError:
                    pass
            else:
                logger.warning(
                    "class_names attribute is deprecated. Please define a class_names method"
                )
                return user_model.class_names
        logger.debug("class_names is not implemented")
        n_targets = predictions.shape[1]
        return ["t:{}".format(i) for i in range(n_targets)]
    else:
        return []


def client_predict(
    user_model: SeldonComponent,
    features: Union[np.ndarray, str, bytes],
    feature_names: Iterable[str],
    **kwargs: Dict,
) -> SeldonResponse:
    """
    Get prediction from user model

    Parameters
    ----------
    user_model
       A seldon user model
    features
       The data payload
    feature_names
       The feature names in the payload
    kwargs
       Optional keyword arguments
    Returns
    -------
       A prediction from the user model
    """
    if hasattr(user_model, "predict"):
        try:
            try:
                client_response = user_model.predict(features, feature_names, **kwargs)
            except TypeError:
                client_response = user_model.predict(features, feature_names)
            return SeldonResponse.create(client_response)
        except SeldonNotImplementedError:
            pass
    logger.debug("predict is not implemented")
    return SeldonResponse.create([])


def client_transform_input(
    user_model: SeldonComponent,
    features: Union[np.ndarray, str, bytes],
    feature_names: Iterable[str],
    **kwargs: Dict,
) -> SeldonResponse:
    """
    Transform data with user model

    Parameters
    ----------
    user_model
       A Seldon user model
    features
       Data payload
    feature_names
       Data payload column names
    kwargs
       Optional keyword args

    Returns
    -------
       Transformed data

    """
    if hasattr(user_model, "transform_input"):
        try:
            try:
                client_response = user_model.transform_input(
                    features, feature_names, **kwargs
                )
            except TypeError:
                client_response = user_model.transform_input(features, feature_names)
            return SeldonResponse.create(client_response)
        except SeldonNotImplementedError:
            pass
    logger.debug("transform_input is not implemented")
    return SeldonResponse.create(features)


def client_transform_output(
    user_model: SeldonComponent,
    features: Union[np.ndarray, str, bytes],
    feature_names: Iterable[str],
    **kwargs: Dict,
) -> SeldonResponse:
    """
    Transform output

    Parameters
    ----------
    user_model
       A Seldon user model
    features
       Data payload
    feature_names
       Data payload column names
    kwargs
       Optional keyword args
    Returns
    -------
       Transformed data

    """
    if hasattr(user_model, "transform_output"):
        try:
            try:
                client_response = user_model.transform_output(
                    features, feature_names, **kwargs
                )
            except TypeError:
                client_response = user_model.transform_output(features, feature_names)
            return SeldonResponse.create(client_response)
        except SeldonNotImplementedError:
            pass
    logger.debug("transform_output is not implemented")
    return SeldonResponse.create(features)


def client_custom_metrics(
    user_model: SeldonComponent,
    seldon_metrics: SeldonMetrics,
    method: str,
    runtime_metrics: List[Dict] = [],
) -> List[Dict]:
    """
    Get custom metrics for client and update SeldonMetrics.

    This function will return empty list if INCLUDE_METRICS_IN_CLIENT_RESPONSE environmental
    variable is NOT set to "true" or "True".

    Parameters
    ----------
    user_model
       A Seldon user model
    seldon_metrics
        A SeldonMetrics instance
    method:
        tag of a method that collected the metrics
    runtime_metrics:
        metrics that were defined on runtime
    Returns
    -------
       A list of custom metrics

    """
    if not validate_metrics(runtime_metrics):
        raise SeldonMicroserviceException(
            f"Bad metric created during request: {json.dumps(runtime_metrics)}",
            status_code=500,
            reason="MICROSERVICE_BAD_METRIC",
        )
    seldon_metrics.update(runtime_metrics, method)

    if hasattr(user_model, "metrics"):
        try:
            metrics = user_model.metrics()
            if not validate_metrics(metrics):
                raise SeldonMicroserviceException(
                    f"Bad metric created during request: {json.dumps(metrics)}",
                    status_code=500,
                    reason="MICROSERVICE_BAD_METRIC",
                )

            seldon_metrics.update(metrics, method)
            if INCLUDE_METRICS_IN_CLIENT_RESPONSE:
                return metrics + runtime_metrics
            else:
                return []
        except SeldonNotImplementedError:
            pass
    logger.debug("custom_metrics is not implemented")
    if INCLUDE_METRICS_IN_CLIENT_RESPONSE:
        return runtime_metrics
    else:
        return []


def client_feature_names(
    user_model: SeldonComponent, original: Iterable[str]
) -> Iterable[str]:
    """
    Get feature names for user model

    Parameters
    ----------
    user_model
       A Seldon user model
    original
       Original feature names
    Returns
    -------
       A list if feature names
    """
    if hasattr(user_model, "feature_names"):
        try:
            return user_model.feature_names()
        except SeldonNotImplementedError:
            pass
    logger.debug("feature_names is not implemented")
    return original


def client_send_feedback(
    user_model: SeldonComponent,
    features: Union[np.ndarray, str, bytes],
    feature_names: Iterable[str],
    reward: float,
    truth: Union[np.ndarray, str, bytes],
    routing: Union[int, None],
) -> SeldonResponse:
    """
    Feedback to user model

    Parameters
    ----------
    user_model
       A Seldon user model
    features
       A payload
    feature_names
       Payload column names
    reward
       Reward
    truth
       True outcome
    routing
       Optional routing

    Returns
    -------
       Optional payload

    """
    if hasattr(user_model, "send_feedback"):
        try:
            client_response = user_model.send_feedback(
                features, feature_names, reward, truth, routing=routing
            )
            return SeldonResponse.create(client_response)
        except SeldonNotImplementedError:
            pass
    logger.debug("send_feedback is not implemented")
    return SeldonResponse.create(None)


def client_route(
    user_model: SeldonComponent,
    features: Union[np.ndarray, str, bytes],
    feature_names: Iterable[str],
    **kwargs: Dict,
) -> SeldonResponse:
    """
    Get routing from user model

    Parameters
    ----------
    user_model
       A Seldon user model
    features
       Payload
    feature_names
       Columns for payload

    Returns
    -------
       Routing index for one of children
    """
    if hasattr(user_model, "route"):
        try:
            client_response = user_model.route(features, feature_names, **kwargs)
        except TypeError:
            client_response = user_model.route(features, feature_names)
        return SeldonResponse.create(client_response)
    else:
        raise SeldonNotImplementedError("Route not defined")


def client_aggregate(
    user_model: SeldonComponent,
    features_list: List[Union[np.ndarray, str, bytes]],
    feature_names_list: List,
) -> SeldonResponse:
    """
    Aggregate payloads

    Parameters
    ----------
    user_model
       A Seldon user model
    features_list
       A list of payloads
    feature_names_list
       Column names for payloads
    Returns
    -------
       An aggregated payload
    """
    if hasattr(user_model, "aggregate"):
        client_response = user_model.aggregate(features_list, feature_names_list)
        return SeldonResponse.create(client_response)
    else:
        raise SeldonNotImplementedError("Aggregate not defined")


def client_health_status(
    user_model: SeldonComponent,
) -> Union[np.ndarray, List, str, bytes]:
    """
    Perform a health check

    Parameters
    ----------
    user_model
       A Seldon user model
    Returns
    -------
       Health check results
    """
    if hasattr(user_model, "health_status"):
        try:
            return user_model.health_status()
        except SeldonNotImplementedError:
            return "not implemented - assuming healthy"
    else:
        return "healthy"



import json
import logging
from unittest.mock import patch

import pytest
import yaml
from google.protobuf import json_format

from seldon_core.metadata import SeldonInvalidMetadataError, validate_model_metadata
from seldon_core.metrics import SeldonMetrics
from seldon_core.proto import prediction_pb2
from seldon_core.utils import json_to_seldon_model_metadata
from seldon_core.wrapper import SeldonModelGRPC, get_rest_microservice

# test default values


def test_default_values():
    meta = validate_model_metadata({})
    assert meta == {
        "name": "",
        "versions": [""],
        "platform": "",
        "inputs": [],
        "outputs": [],
        "custom": {},
    }


def test_default_values_with_env():
    with patch("seldon_core.metadata.MODEL_IMAGE", "seldonio/sklearn-iris:0.1"):
        meta = validate_model_metadata({})
    assert meta == {
        "name": "seldonio/sklearn-iris",
        "versions": ["0.1"],
        "platform": "",
        "inputs": [],
        "outputs": [],
        "custom": {},
    }


def test_default_values_with_colon_in_env():
    with patch("seldon_core.metadata.MODEL_IMAGE", "localhost:32000/sklearn-iris:0.1"):
        meta = validate_model_metadata({})
    assert meta == {
        "name": "localhost:32000/sklearn-iris",
        "versions": ["0.1"],
        "platform": "",
        "inputs": [],
        "outputs": [],
        "custom": {},
    }


# V1 meta tests


@pytest.mark.parametrize(
    "messagetype",
    ["tensor", "ndarray", "tftensor", "jsonData", "binData", "strData", "custom_data"],
)
def test_v1_only_messagetype(messagetype):
    data = f"""
        name: my-model-name
        versions: [ my-model-version-01 ]
        platform: seldon
        inputs:
        - messagetype: {messagetype}
        outputs:
        - messagetype: {messagetype}
        custom:
          tag-key: tag-value
    """
    meta_json = validate_model_metadata(yaml.safe_load(data))
    meta_proto = json_to_seldon_model_metadata(meta_json)
    assert meta_json == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [{"messagetype": f"{messagetype}"}],
        "outputs": [{"messagetype": f"{messagetype}"}],
        "custom": {"tag-key": "tag-value"},
    }
    assert json.loads(json_format.MessageToJson(meta_proto)) == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [{"messagetype": f"{messagetype}"}],
        "outputs": [{"messagetype": f"{messagetype}"}],
        "custom": {"tag-key": "tag-value"},
    }


def test_v1_mixed_multiple_inputs():
    data = """
        name: my-model-name
        versions: [ my-model-version-01 ]
        platform: seldon
        inputs:
        - messagetype: "tensor"
          schema:
            names: [a, b, c, d]
            shape: [ 2, 2 ]
        - messagetype: jsonData
        outputs:
        - messagetype: "binData"
        custom:
          tag-key: tag-value
    """
    meta_json = validate_model_metadata(yaml.safe_load(data))
    meta_proto = json_to_seldon_model_metadata(meta_json)
    assert meta_json == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [
            {
                "messagetype": "tensor",
                "schema": {"names": ["a", "b", "c", "d"], "shape": [2, 2]},
            },
            {"messagetype": "jsonData"},
        ],
        "outputs": [{"messagetype": "binData"}],
        "custom": {"tag-key": "tag-value"},
    }
    assert json.loads(json_format.MessageToJson(meta_proto)) == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [
            {
                "messagetype": "tensor",
                "schema": {"names": ["a", "b", "c", "d"], "shape": [2, 2]},
            },
            {"messagetype": "jsonData"},
        ],
        "outputs": [{"messagetype": "binData"}],
        "custom": {"tag-key": "tag-value"},
    }


@pytest.mark.parametrize(
    "messagetype", ["tensor", "ndarray", "tftensor", "custom_data"]
)
def test_v1_array(messagetype):
    data = f"""
        name: my-model-name
        versions: [ my-model-version-01 ]
        platform: seldon
        inputs:
        - messagetype: {messagetype}
          schema:
            names: [a, b, c, d]
            shape: [ 2, 2 ]
        outputs:
        - messagetype: {messagetype}
          schema:
            shape: [ 1 ]
        custom:
          tag-key: tag-value
    """
    meta_json = validate_model_metadata(yaml.safe_load(data))
    meta_proto = json_to_seldon_model_metadata(meta_json)
    assert meta_json == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [
            {
                "messagetype": f"{messagetype}",
                "schema": {"names": ["a", "b", "c", "d"], "shape": [2, 2]},
            }
        ],
        "outputs": [{"messagetype": f"{messagetype}", "schema": {"shape": [1]}}],
        "custom": {"tag-key": "tag-value"},
    }
    assert json.loads(json_format.MessageToJson(meta_proto)) == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [
            {
                "messagetype": f"{messagetype}",
                "schema": {"names": ["a", "b", "c", "d"], "shape": [2.0, 2.0]},
            }
        ],
        "outputs": [{"messagetype": f"{messagetype}", "schema": {"shape": [1.0]}}],
        "custom": {"tag-key": "tag-value"},
    }


def test_v1_json_with_schema():
    data = """
        name: my-model-name
        versions: [ my-model-version-01 ]
        platform: seldon
        inputs:
        - messagetype: jsonData
          schema:
              type: object
              properties:
                  names:
                      type: array
                      items:
                          type: string
                  data:
                    type: array
                    items:
                        type: number
                        format: double
        outputs:
        - messagetype: jsonData
        custom:
          tag-key: tag-value
    """
    meta_json = validate_model_metadata(yaml.safe_load(data))
    meta_proto = json_to_seldon_model_metadata(meta_json)
    assert meta_json == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [
            {
                "messagetype": "jsonData",
                "schema": {
                    "type": "object",
                    "properties": {
                        "names": {"type": "array", "items": {"type": "string"}},
                        "data": {
                            "type": "array",
                            "items": {"type": "number", "format": "double"},
                        },
                    },
                },
            }
        ],
        "outputs": [{"messagetype": "jsonData"}],
        "custom": {"tag-key": "tag-value"},
    }
    assert json.loads(json_format.MessageToJson(meta_proto)) == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [
            {
                "messagetype": "jsonData",
                "schema": {
                    "properties": {
                        "names": {"items": {"type": "string"}, "type": "array"},
                        "data": {
                            "type": "array",
                            "items": {"type": "number", "format": "double"},
                        },
                    },
                    "type": "object",
                },
            }
        ],
        "outputs": [{"messagetype": "jsonData"}],
        "custom": {"tag-key": "tag-value"},
    }


# V1 meta tests (failures)


@pytest.mark.parametrize(
    "invalid_input",
    [
        {},  # no such schema
        {"messagetype": "mytype"},  # to such valid schema
        {"messagetype": "array"},  # fails because shape is missing
        {"messagetype": "array", "shape": "1, 2"},  # shape is wrong type
        {"messagetype": "array", "shape": "1, 2"},  # shape is wrong type
        {"messagetype": "array", "shape": [2, 2], "invalid": "field"},
        {"messagetype": "jsonData", "invalid": "field"},
        {"messagetype": "jsonData", "schema": "some string"},  # schema should be dict
        {"messagetype": "strData", "invalid": "field"},
        {"messagetype": "binData", "invalid": "field"},
    ],
)
def test_v1_invalid_inputs(invalid_input):
    valid_base = {
        "inputs": [{"messagetype": "strData"}],
        "outputs": [{"messagetype": "strData"}],
    }
    with pytest.raises(SeldonInvalidMetadataError):
        validate_model_metadata({**valid_base, **{"inputs": invalid_input}})
        validate_model_metadata({**valid_base, **{"outputs": invalid_input}})


@pytest.mark.parametrize(
    "messagetype", ["tensor", "ndarray", "tftensor", "binData", "strData"]
)
def test_v1_invalid_schema_fields(messagetype):
    meta = {
        "inputs": [
            {"messagetype": messagetype, "schema": {"custom-field": "custom-def"}}
        ]
    }
    with pytest.raises(SeldonInvalidMetadataError):
        validate_model_metadata(meta)


@pytest.mark.parametrize("messagetype", ["jsonData", "customData"])
def test_v1_valid_custom_schema(messagetype):
    meta = {
        "inputs": [
            {"messagetype": messagetype, "schema": {"custom-field": "custom-def"}}
        ]
    }
    validate_model_metadata(meta)


# V2 meta tests


def test_v2():
    data = """
        name: my-model-name
        versions: [ my-model-version-01 ]
        platform: seldon
        inputs:
        - datatype: BYTES
          name: input
          shape: [ 1, 4 ]
        outputs:
        - datatype: BYTES
          name: output
          shape: [ 3 ]
        custom:
          tag-key: tag-value
    """
    meta_json = validate_model_metadata(yaml.safe_load(data))
    meta_proto = json_to_seldon_model_metadata(meta_json)
    assert meta_json == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [{"datatype": "BYTES", "name": "input", "shape": [1, 4]}],
        "outputs": [{"datatype": "BYTES", "name": "output", "shape": [3]}],
        "custom": {"tag-key": "tag-value"},
    }
    assert json.loads(json_format.MessageToJson(meta_proto)) == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [{"name": "input", "datatype": "BYTES", "shape": ["1", "4"]}],
        "outputs": [{"name": "output", "datatype": "BYTES", "shape": ["3"]}],
        "custom": {"tag-key": "tag-value"},
    }


# mix test


def test_v1_v2_mix():
    data = """
        name: my-model-name
        versions: [ my-model-version-01 ]
        platform: seldon
        inputs:
        - datatype: BYTES
          name: input
          shape: [ 1, 4 ]
        - messagetype: jsonData
        outputs:
        - datatype: BYTES
          name: output
          shape: [ 3 ]
        custom:
          tag-key: tag-value
    """
    meta_json = validate_model_metadata(yaml.safe_load(data))
    meta_proto = json_to_seldon_model_metadata(meta_json)
    assert meta_json == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [
            {"datatype": "BYTES", "name": "input", "shape": [1, 4]},
            {"messagetype": "jsonData"},
        ],
        "outputs": [{"datatype": "BYTES", "name": "output", "shape": [3]}],
        "custom": {"tag-key": "tag-value"},
    }
    assert json.loads(json_format.MessageToJson(meta_proto)) == {
        "name": "my-model-name",
        "versions": ["my-model-version-01"],
        "platform": "seldon",
        "inputs": [
            {"name": "input", "datatype": "BYTES", "shape": ["1", "4"]},
            {"messagetype": "jsonData"},
        ],
        "outputs": [{"name": "output", "datatype": "BYTES", "shape": ["3"]}],
        "custom": {"tag-key": "tag-value"},
    }


# validate_model_metadata tests block


def test_validate_model_metadata():
    meta = {
        "name": "my-model-name",
        "versions": ["model-version"],
        "platform": "model-platform",
        "inputs": [{"name": "input", "datatype": "BYTES", "shape": [1]}],
        "outputs": [{"name": "output", "datatype": "BYTES", "shape": [1]}],
        "custom": {"tag-key": "tag-value"},
    }
    with patch("seldon_core.metadata.MODEL_IMAGE", None):
        assert meta == validate_model_metadata(meta)


def test_validate_model_metadata_with_env():
    meta = {
        "name": "my-model-name",
        "versions": ["model-version"],
        "platform": "model-platform",
        "inputs": [{"name": "input", "datatype": "BYTES", "shape": [1]}],
        "outputs": [{"name": "output", "datatype": "BYTES", "shape": [1]}],
        "custom": {"tag-key": "tag-value"},
    }
    with patch("seldon_core.metadata.MODEL_IMAGE", "seldonio/sklearn-iris:0.1"):
        assert meta == validate_model_metadata(meta)


def test_validate_model_metadata_with_colon_in_env():
    meta = {
        "name": "my-model-name",
        "versions": ["model-version"],
        "platform": "model-platform",
        "inputs": [{"name": "input", "datatype": "BYTES", "shape": [1]}],
        "outputs": [{"name": "output", "datatype": "BYTES", "shape": [1]}],
        "custom": {"tag-key": "tag-value"},
    }
    with patch("seldon_core.metadata.MODEL_IMAGE", "localhost:32000/sklearn-iris:0.1"):
        assert meta == validate_model_metadata(meta)


@pytest.mark.parametrize("invalid_versions", ["v1", [1], "[v]", "[1]", 1, 1.1])
def test_validate_model_metadata_wrong_versions(invalid_versions):
    with pytest.raises(SeldonInvalidMetadataError):
        validate_model_metadata({"versions": invalid_versions})


@pytest.mark.parametrize(
    "invalid_tensor",
    [
        {"name": "tensor-name", "datatype": "data-type", "shape": [1, 2]},
        "some string",
        ["some string in array"],
    ],
)
def test_validate_model_metadata_wrong_inputs_outputs(invalid_tensor):
    # Note: this could not be combined with test_model_metadata_wrong_inputs_outputs as
    # second input there is a valid input here
    with pytest.raises(SeldonInvalidMetadataError):
        validate_model_metadata({"inputs": invalid_tensor})
    with pytest.raises(SeldonInvalidMetadataError):
        validate_model_metadata({"outputs": invalid_tensor})


# Microservice tests block

yaml_meta = """
---
name: test-name-env
versions: [ test-version ]
platform: test-platform
inputs:
- datatype: BYTES
  name: input
  shape: [ 1, 2 ]
outputs:
- datatype: BYTES
  name: output
  shape: [ 1, 2, 3 ]
custom:
  tag-key: tag-value
"""


json_meta = """
{
    "name": "test_name_json",
    "versions": ["test-version"],
    "platform": "seldon",
    "inputs": [{"name": "input", "datatype": "BYTES", "shape": [1, 2]}],
    "outputs": [{"name": "output", "datatype": "BYTES", "shape": [1, 2, 3]}],
    "custom": {"tag-key": "tag-value"},
}
"""


class UserObject:
    METADATA_RESPONSE = {
        "name": "my-model-name",
        "versions": ["model-version"],
        "platform": "model-platform",
        "inputs": [{"name": "input", "datatype": "BYTES", "shape": [1]}],
        "outputs": [{"name": "output", "datatype": "BYTES", "shape": [1]}],
        "custom": {"tag-key": "tag-value"},
    }

    def predict(self, X, features_names):
        logging.info("Predict called")
        return X

    def init_metadata(self):
        logging.info("init_metadata called")
        return self.METADATA_RESPONSE


def test_model_metadata_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()

    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()

    rv = client.get('/predict?json={"data": {"names": ["input"], "ndarray": ["data"]}}')
    assert rv.status_code == 200
    assert json.loads(rv.data)["data"]["ndarray"] == ["data"]

    rv = client.get("/metadata")
    assert rv.status_code == 200


def test_model_metadata_ok_grpc():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()

    app = SeldonModelGRPC(user_object, seldon_metrics)
    resp = app.Metadata(None, None)
    assert json.loads(json_format.MessageToJson(resp)) == {
        "name": "my-model-name",
        "versions": ["model-version"],
        "platform": "model-platform",
        "inputs": [{"name": "input", "datatype": "BYTES", "shape": ["1"]}],
        "outputs": [{"name": "output", "datatype": "BYTES", "shape": ["1"]}],
        "custom": {"tag-key": "tag-value"},
    }


@pytest.mark.parametrize("env_value", [json_meta, yaml_meta])
def test_model_metadata_value_in_env(env_value):
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()

    # it is enough to only patch call to get_rest_microservice as it caches the metadata
    with patch("os.environ", {"MODEL_METADATA": env_value}):
        app = get_rest_microservice(user_object, seldon_metrics)

    client = app.test_client()

    rv = client.get('/predict?json={"data": {"names": ["input"], "ndarray": ["data"]}}')
    assert rv.status_code == 200
    assert json.loads(rv.data)["data"]["ndarray"] == ["data"]

    rv = client.get("/metadata")
    assert rv.status_code == 200
    assert json.loads(rv.data) == {
        **UserObject.METADATA_RESPONSE,
        **yaml.safe_load(env_value),
    }


def test_model_metadata_invalid_user_definition():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()

    # Break UserObject class to provide invalid init_metadata
    user_object.METADATA_RESPONSE["versions"] = "string-not-array-of-strings"

    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()

    rv = client.get('/predict?json={"data": {"names": ["input"], "ndarray": ["data"]}}')
    assert rv.status_code == 200
    assert json.loads(rv.data)["data"]["ndarray"] == ["data"]

    rv = client.get("/metadata")
    assert rv.status_code == 500
    assert json.loads(rv.data) == {
        "status": {
            "code": -1,
            "info": "Model metadata unavailable",
            "reason": "MICROSERVICE_BAD_METADATA",
            "status": 1,
        }
    }



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tensorflow/core/framework/resource_handle.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'tensorflow/core/framework/resource_handle.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import tensor_shape_pb2 as tensorflow_dot_core_dot_framework_dot_tensor__shape__pb2
from . import types_pb2 as tensorflow_dot_core_dot_framework_dot_types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/tensorflow/core/framework/resource_handle.proto\x12\ntensorflow\x1a,tensorflow/core/framework/tensor_shape.proto\x1a%tensorflow/core/framework/types.proto\"\xa5\x02\n\x13ResourceHandleProto\x12\x0e\n\x06\x64\x65vice\x18\x01 \x01(\t\x12\x11\n\tcontainer\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x11\n\thash_code\x18\x04 \x01(\x04\x12\x17\n\x0fmaybe_type_name\x18\x05 \x01(\t\x12H\n\x11\x64types_and_shapes\x18\x06 \x03(\x0b\x32-.tensorflow.ResourceHandleProto.DtypeAndShape\x1a\x61\n\rDtypeAndShape\x12#\n\x05\x64type\x18\x01 \x01(\x0e\x32\x14.tensorflow.DataType\x12+\n\x05shape\x18\x02 \x01(\x0b\x32\x1c.tensorflow.TensorShapeProtoJ\x04\x08\x07\x10\x08\x42\x87\x01\n\x18org.tensorflow.frameworkB\x0eResourceHandleP\x01ZVgithub.com/tensorflow/tensorflow/tensorflow/go/core/framework/resource_handle_go_proto\xf8\x01\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tensorflow.core.framework.resource_handle_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030org.tensorflow.frameworkB\016ResourceHandleP\001ZVgithub.com/tensorflow/tensorflow/tensorflow/go/core/framework/resource_handle_go_proto\370\001\001'
  _globals['_RESOURCEHANDLEPROTO']._serialized_start=149
  _globals['_RESOURCEHANDLEPROTO']._serialized_end=442
  _globals['_RESOURCEHANDLEPROTO_DTYPEANDSHAPE']._serialized_start=339
  _globals['_RESOURCEHANDLEPROTO_DTYPEANDSHAPE']._serialized_end=436
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: proto/prediction.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'proto/prediction.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from seldon_core.proto.tensorflow.core.framework import tensor_pb2 as tensorflow_dot_core_dot_framework_dot_tensor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16proto/prediction.proto\x12\rseldon.protos\x1a\x19google/protobuf/any.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a&tensorflow/core/framework/tensor.proto\"\x91\x02\n\rSeldonMessage\x12%\n\x06status\x18\x01 \x01(\x0b\x32\x15.seldon.protos.Status\x12!\n\x04meta\x18\x02 \x01(\x0b\x32\x13.seldon.protos.Meta\x12*\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\x1a.seldon.protos.DefaultDataH\x00\x12\x11\n\x07\x62inData\x18\x04 \x01(\x0cH\x00\x12\x11\n\x07strData\x18\x05 \x01(\tH\x00\x12*\n\x08jsonData\x18\x06 \x01(\x0b\x32\x16.google.protobuf.ValueH\x00\x12*\n\ncustomData\x18\x07 \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x42\x0c\n\ndata_oneof\"\xaf\x01\n\x0b\x44\x65\x66\x61ultData\x12\r\n\x05names\x18\x01 \x03(\t\x12\'\n\x06tensor\x18\x02 \x01(\x0b\x32\x15.seldon.protos.TensorH\x00\x12-\n\x07ndarray\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.ListValueH\x00\x12+\n\x08tftensor\x18\x04 \x01(\x0b\x32\x17.tensorflow.TensorProtoH\x00\x42\x0c\n\ndata_oneof\"/\n\x06Tensor\x12\x11\n\x05shape\x18\x01 \x03(\x05\x42\x02\x10\x01\x12\x12\n\x06values\x18\x02 \x03(\x01\x42\x02\x10\x01\"\x80\x03\n\x04Meta\x12\x0c\n\x04puid\x18\x01 \x01(\t\x12+\n\x04tags\x18\x02 \x03(\x0b\x32\x1d.seldon.protos.Meta.TagsEntry\x12\x31\n\x07routing\x18\x03 \x03(\x0b\x32 .seldon.protos.Meta.RoutingEntry\x12\x39\n\x0brequestPath\x18\x04 \x03(\x0b\x32$.seldon.protos.Meta.RequestPathEntry\x12&\n\x07metrics\x18\x05 \x03(\x0b\x32\x15.seldon.protos.Metric\x1a\x43\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12%\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.Value:\x02\x38\x01\x1a.\n\x0cRoutingEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a\x32\n\x10RequestPathEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xe1\x01\n\x06Metric\x12\x0b\n\x03key\x18\x01 \x01(\t\x12.\n\x04type\x18\x02 \x01(\x0e\x32 .seldon.protos.Metric.MetricType\x12\r\n\x05value\x18\x03 \x01(\x02\x12-\n\x04tags\x18\x04 \x03(\x0b\x32\x1f.seldon.protos.Metric.TagsEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"/\n\nMetricType\x12\x0b\n\x07\x43OUNTER\x10\x00\x12\t\n\x05GAUGE\x10\x01\x12\t\n\x05TIMER\x10\x02\"I\n\x11SeldonMessageList\x12\x34\n\x0eseldonMessages\x18\x01 \x03(\x0b\x32\x1c.seldon.protos.SeldonMessage\"\x8e\x01\n\x06Status\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0c\n\x04info\x18\x02 \x01(\t\x12\x0e\n\x06reason\x18\x03 \x01(\t\x12\x30\n\x06status\x18\x04 \x01(\x0e\x32 .seldon.protos.Status.StatusFlag\"&\n\nStatusFlag\x12\x0b\n\x07SUCCESS\x10\x00\x12\x0b\n\x07\x46\x41ILURE\x10\x01\"\xa6\x01\n\x08\x46\x65\x65\x64\x62\x61\x63k\x12-\n\x07request\x18\x01 \x01(\x0b\x32\x1c.seldon.protos.SeldonMessage\x12.\n\x08response\x18\x02 \x01(\x0b\x32\x1c.seldon.protos.SeldonMessage\x12\x0e\n\x06reward\x18\x03 \x01(\x02\x12+\n\x05truth\x18\x04 \x01(\x0b\x32\x1c.seldon.protos.SeldonMessage\"p\n\x0fRequestResponse\x12-\n\x07request\x18\x01 \x01(\x0b\x32\x1c.seldon.protos.SeldonMessage\x12.\n\x08response\x18\x02 \x01(\x0b\x32\x1c.seldon.protos.SeldonMessage\"*\n\x1aSeldonModelMetadataRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x83\x01\n\x15SeldonMessageMetadata\x12\x13\n\x0bmessagetype\x18\x01 \x01(\t\x12&\n\x06schema\x18\x02 \x01(\x0b\x32\x16.google.protobuf.Value\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x04 \x01(\t\x12\r\n\x05shape\x18\x05 \x03(\x03\"\xa3\x02\n\x13SeldonModelMetadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08versions\x18\x02 \x03(\t\x12\x10\n\x08platform\x18\x03 \x01(\t\x12\x34\n\x06inputs\x18\x04 \x03(\x0b\x32$.seldon.protos.SeldonMessageMetadata\x12\x35\n\x07outputs\x18\x05 \x03(\x0b\x32$.seldon.protos.SeldonMessageMetadata\x12>\n\x06\x63ustom\x18\x06 \x03(\x0b\x32..seldon.protos.SeldonModelMetadata.CustomEntry\x1a-\n\x0b\x43ustomEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa3\x02\n\x13SeldonGraphMetadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12>\n\x06models\x18\x02 \x03(\x0b\x32..seldon.protos.SeldonGraphMetadata.ModelsEntry\x12\x34\n\x06inputs\x18\x03 \x03(\x0b\x32$.seldon.protos.SeldonMessageMetadata\x12\x35\n\x07outputs\x18\x04 \x03(\x0b\x32$.seldon.protos.SeldonMessageMetadata\x1aQ\n\x0bModelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x31\n\x05value\x18\x02 \x01(\x0b\x32\".seldon.protos.SeldonModelMetadata:\x02\x38\x01\x32\x89\x03\n\x07Generic\x12N\n\x0eTransformInput\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12O\n\x0fTransformOutput\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12\x45\n\x05Route\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12M\n\tAggregate\x12 .seldon.protos.SeldonMessageList\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12G\n\x0cSendFeedback\x12\x17.seldon.protos.Feedback\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x32\xe3\x01\n\x05Model\x12G\n\x07Predict\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12G\n\x0cSendFeedback\x12\x17.seldon.protos.Feedback\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12H\n\x08Metadata\x12\x16.google.protobuf.Empty\x1a\".seldon.protos.SeldonModelMetadata\"\x00\x32\x98\x01\n\x06Router\x12\x45\n\x05Route\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12G\n\x0cSendFeedback\x12\x17.seldon.protos.Feedback\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x32]\n\x0bTransformer\x12N\n\x0eTransformInput\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x32\x64\n\x11OutputTransformer\x12O\n\x0fTransformOutput\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x32Y\n\x08\x43ombiner\x12M\n\tAggregate\x12 .seldon.protos.SeldonMessageList\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x32\xcb\x02\n\x06Seldon\x12G\n\x07Predict\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12G\n\x0cSendFeedback\x12\x17.seldon.protos.Feedback\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12`\n\rModelMetadata\x12).seldon.protos.SeldonModelMetadataRequest\x1a\".seldon.protos.SeldonModelMetadata\"\x00\x12M\n\rGraphMetadata\x12\x16.google.protobuf.Empty\x1a\".seldon.protos.SeldonGraphMetadata\"\x00\x42h\n\x10io.seldon.protosB\x10PredictionProtosZBgithub.com/seldonio/seldon-core/incubating/wrappers/s2i/go/pkg/apib\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'proto.prediction_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020io.seldon.protosB\020PredictionProtosZBgithub.com/seldonio/seldon-core/incubating/wrappers/s2i/go/pkg/api'
  _globals['_TENSOR'].fields_by_name['shape']._loaded_options = None
  _globals['_TENSOR'].fields_by_name['shape']._serialized_options = b'\020\001'
  _globals['_TENSOR'].fields_by_name['values']._loaded_options = None
  _globals['_TENSOR'].fields_by_name['values']._serialized_options = b'\020\001'
  _globals['_META_TAGSENTRY']._loaded_options = None
  _globals['_META_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_META_ROUTINGENTRY']._loaded_options = None
  _globals['_META_ROUTINGENTRY']._serialized_options = b'8\001'
  _globals['_META_REQUESTPATHENTRY']._loaded_options = None
  _globals['_META_REQUESTPATHENTRY']._serialized_options = b'8\001'
  _globals['_METRIC_TAGSENTRY']._loaded_options = None
  _globals['_METRIC_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_SELDONMODELMETADATA_CUSTOMENTRY']._loaded_options = None
  _globals['_SELDONMODELMETADATA_CUSTOMENTRY']._serialized_options = b'8\001'
  _globals['_SELDONGRAPHMETADATA_MODELSENTRY']._loaded_options = None
  _globals['_SELDONGRAPHMETADATA_MODELSENTRY']._serialized_options = b'8\001'
  _globals['_SELDONMESSAGE']._serialized_start=168
  _globals['_SELDONMESSAGE']._serialized_end=441
  _globals['_DEFAULTDATA']._serialized_start=444
  _globals['_DEFAULTDATA']._serialized_end=619
  _globals['_TENSOR']._serialized_start=621
  _globals['_TENSOR']._serialized_end=668
  _globals['_META']._serialized_start=671
  _globals['_META']._serialized_end=1055
  _globals['_META_TAGSENTRY']._serialized_start=888
  _globals['_META_TAGSENTRY']._serialized_end=955
  _globals['_META_ROUTINGENTRY']._serialized_start=957
  _globals['_META_ROUTINGENTRY']._serialized_end=1003
  _globals['_META_REQUESTPATHENTRY']._serialized_start=1005
  _globals['_META_REQUESTPATHENTRY']._serialized_end=1055
  _globals['_METRIC']._serialized_start=1058
  _globals['_METRIC']._serialized_end=1283
  _globals['_METRIC_TAGSENTRY']._serialized_start=1191
  _globals['_METRIC_TAGSENTRY']._serialized_end=1234
  _globals['_METRIC_METRICTYPE']._serialized_start=1236
  _globals['_METRIC_METRICTYPE']._serialized_end=1283
  _globals['_SELDONMESSAGELIST']._serialized_start=1285
  _globals['_SELDONMESSAGELIST']._serialized_end=1358
  _globals['_STATUS']._serialized_start=1361
  _globals['_STATUS']._serialized_end=1503
  _globals['_STATUS_STATUSFLAG']._serialized_start=1465
  _globals['_STATUS_STATUSFLAG']._serialized_end=1503
  _globals['_FEEDBACK']._serialized_start=1506
  _globals['_FEEDBACK']._serialized_end=1672
  _globals['_REQUESTRESPONSE']._serialized_start=1674
  _globals['_REQUESTRESPONSE']._serialized_end=1786
  _globals['_SELDONMODELMETADATAREQUEST']._serialized_start=1788
  _globals['_SELDONMODELMETADATAREQUEST']._serialized_end=1830
  _globals['_SELDONMESSAGEMETADATA']._serialized_start=1833
  _globals['_SELDONMESSAGEMETADATA']._serialized_end=1964
  _globals['_SELDONMODELMETADATA']._serialized_start=1967
  _globals['_SELDONMODELMETADATA']._serialized_end=2258
  _globals['_SELDONMODELMETADATA_CUSTOMENTRY']._serialized_start=2213
  _globals['_SELDONMODELMETADATA_CUSTOMENTRY']._serialized_end=2258
  _globals['_SELDONGRAPHMETADATA']._serialized_start=2261
  _globals['_SELDONGRAPHMETADATA']._serialized_end=2552
  _globals['_SELDONGRAPHMETADATA_MODELSENTRY']._serialized_start=2471
  _globals['_SELDONGRAPHMETADATA_MODELSENTRY']._serialized_end=2552
  _globals['_GENERIC']._serialized_start=2555
  _globals['_GENERIC']._serialized_end=2948
  _globals['_MODEL']._serialized_start=2951
  _globals['_MODEL']._serialized_end=3178
  _globals['_ROUTER']._serialized_start=3181
  _globals['_ROUTER']._serialized_end=3333
  _globals['_TRANSFORMER']._serialized_start=3335
  _globals['_TRANSFORMER']._serialized_end=3428
  _globals['_OUTPUTTRANSFORMER']._serialized_start=3430
  _globals['_OUTPUTTRANSFORMER']._serialized_end=3530
  _globals['_COMBINER']._serialized_start=3532
  _globals['_COMBINER']._serialized_end=3621
  _globals['_SELDON']._serialized_start=3624
  _globals['_SELDON']._serialized_end=3955
# @@protoc_insertion_point(module_scope)



import logging
import os
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import yaml
from google.protobuf import json_format

from seldon_core.flask_utils import SeldonMicroserviceException
from seldon_core.metadata import SeldonInvalidMetadataError, validate_model_metadata
from seldon_core.metrics import (
    AGGREGATE_METRIC_METHOD_TAG,
    FEEDBACK_METRIC_METHOD_TAG,
    HEALTH_METRIC_METHOD_TAG,
    INPUT_TRANSFORM_METRIC_METHOD_TAG,
    OUTPUT_TRANSFORM_METRIC_METHOD_TAG,
    PREDICT_METRIC_METHOD_TAG,
    ROUTER_METRIC_METHOD_TAG,
    SeldonMetrics,
)
from seldon_core.proto import prediction_pb2
from seldon_core.user_model import (
    INCLUDE_METRICS_IN_CLIENT_RESPONSE,
    SeldonNotImplementedError,
    SeldonResponse,
    client_aggregate,
    client_custom_metrics,
    client_health_status,
    client_predict,
    client_route,
    client_send_feedback,
    client_transform_input,
    client_transform_output,
)
from seldon_core.utils import (
    construct_response,
    construct_response_json,
    extract_feedback_request_parts,
    extract_request_parts,
    extract_request_parts_json,
    getenv_as_bool,
    json_to_seldon_message,
    seldon_message_to_json,
)

logger = logging.getLogger(__name__)


def handle_raw_custom_metrics(
    msg: Union[prediction_pb2.SeldonMessage, Dict],
    seldon_metrics: SeldonMetrics,
    is_proto: bool,
    method: str,
):
    """
    Update SeldonMetrics object with custom metrics from raw methods.
    If INCLUDE_METRICS_IN_CLIENT_RESPONSE environmental variable is set to "true"
    metrics will be dropped from msg.
    """
    metrics = []
    if is_proto:
        metrics = seldon_message_to_json(msg.meta).get("metrics", [])
        if metrics and not INCLUDE_METRICS_IN_CLIENT_RESPONSE:
            del msg.meta.metrics[:]
    elif isinstance(msg, dict):
        metrics = msg.get("meta", {}).get("metrics", [])
        if metrics and not INCLUDE_METRICS_IN_CLIENT_RESPONSE:
            del msg["meta"]["metrics"]
    seldon_metrics.update(metrics, method)


def predict(
    user_model: Any,
    request: Union[prediction_pb2.SeldonMessage, List, Dict, bytes],
    seldon_metrics: SeldonMetrics,
) -> Union[prediction_pb2.SeldonMessage, List, Dict, bytes]:
    """
    Call the user model to get a prediction and package the response

    Parameters
    ----------
    user_model
       User defined class instance
    request
       The incoming request

    Returns
    -------
      The prediction
    """
    # TODO: Find a way to choose predict_rest or predict_grpc when payload is
    # not decoded
    is_proto = isinstance(request, prediction_pb2.SeldonMessage)

    if hasattr(user_model, "predict_rest") and not is_proto:
        logger.warning("predict_rest is deprecated. Please use predict_raw")
        return user_model.predict_rest(request)
    elif hasattr(user_model, "predict_grpc") and is_proto:
        logger.warning("predict_grpc is deprecated. Please use predict_raw")
        return user_model.predict_grpc(request)
    else:
        if hasattr(user_model, "predict_raw"):
            try:
                response = user_model.predict_raw(request)
                handle_raw_custom_metrics(
                    response, seldon_metrics, is_proto, PREDICT_METRIC_METHOD_TAG
                )
                return response
            except SeldonNotImplementedError:
                pass

        if is_proto:
            (features, meta, datadef, data_type) = extract_request_parts(request)

            client_response = client_predict(
                user_model, features, datadef.names, meta=meta
            )

            metrics = client_custom_metrics(
                user_model,
                seldon_metrics,
                PREDICT_METRIC_METHOD_TAG,
                client_response.metrics,
            )

            return construct_response(
                user_model,
                False,
                request,
                client_response.data,
                meta,
                metrics,
                client_response.tags,
            )
        else:
            (features, meta, datadef, data_type) = extract_request_parts_json(request)
            class_names = datadef["names"] if datadef and "names" in datadef else []

            client_response = client_predict(
                user_model, features, class_names, meta=meta
            )

            metrics = client_custom_metrics(
                user_model,
                seldon_metrics,
                PREDICT_METRIC_METHOD_TAG,
                client_response.metrics,
            )

            return construct_response_json(
                user_model,
                False,
                request,
                client_response.data,
                meta,
                metrics,
                client_response.tags,
            )


def send_feedback(
    user_model: Any,
    request: prediction_pb2.Feedback,
    predictive_unit_id: str,
    seldon_metrics: SeldonMetrics,
) -> prediction_pb2.SeldonMessage:
    """

    Parameters
    ----------
    user_model
       A Seldon user model
    request
       SeldonMesage proto
    predictive_unit_id
       The ID of the enclosing container predictive unit. Will be taken from environment.

    Returns
    -------

    """
    seldon_metrics.update_reward(request.reward)

    if hasattr(user_model, "send_feedback_rest"):
        logger.warning("send_feedback_rest is deprecated. Please use send_feedback_raw")
        request_json = json_format.MessageToJson(request)
        response_json = user_model.send_feedback_rest(request_json)
        return json_to_seldon_message(response_json)
    elif hasattr(user_model, "send_feedback_grpc"):
        logger.warning("send_feedback_grpc is deprecated. Please use send_feedback_raw")
        response_json = user_model.send_feedback_grpc(request)
        return json_to_seldon_message(response_json)
    else:
        if hasattr(user_model, "send_feedback_raw"):
            try:
                response = user_model.send_feedback_raw(request)
                handle_raw_custom_metrics(
                    response, seldon_metrics, True, FEEDBACK_METRIC_METHOD_TAG
                )
                return response
            except SeldonNotImplementedError:
                pass

        (datadef_request, features, truth, reward) = extract_feedback_request_parts(
            request
        )
        routing = request.response.meta.routing.get(predictive_unit_id)

        client_response = client_send_feedback(
            user_model, features, datadef_request.names, reward, truth, routing
        )

        metrics = client_custom_metrics(
            user_model,
            seldon_metrics,
            FEEDBACK_METRIC_METHOD_TAG,
            client_response.metrics,
        )

        if client_response.data is None:
            client_response.data = np.array([])

        return construct_response(
            user_model,
            False,
            request.request,
            client_response.data,
            None,
            metrics,
            client_response.tags,
        )


def transform_input(
    user_model: Any,
    request: Union[prediction_pb2.SeldonMessage, List, Dict],
    seldon_metrics: SeldonMetrics,
) -> Union[prediction_pb2.SeldonMessage, List, Dict]:
    """

    Parameters
    ----------
    user_model
       User defined class to handle transform input
    request
       The incoming request

    Returns
    -------
       The transformed request

    """
    is_proto = isinstance(request, prediction_pb2.SeldonMessage)

    if hasattr(user_model, "transform_input_rest"):
        logger.warning(
            "transform_input_rest is deprecated. Please use transform_input_raw"
        )
        return user_model.transform_input_rest(request)
    elif hasattr(user_model, "transform_input_grpc"):
        logger.warning(
            "transform_input_grpc is deprecated. Please use transform_input_raw"
        )
        return user_model.transform_input_grpc(request)
    else:
        if hasattr(user_model, "transform_input_raw"):
            try:
                response = user_model.transform_input_raw(request)
                handle_raw_custom_metrics(
                    response,
                    seldon_metrics,
                    is_proto,
                    INPUT_TRANSFORM_METRIC_METHOD_TAG,
                )
                return response
            except SeldonNotImplementedError:
                pass

        if is_proto:
            (features, meta, datadef, data_type) = extract_request_parts(request)

            client_response = client_transform_input(
                user_model, features, datadef.names, meta=meta
            )

            metrics = client_custom_metrics(
                user_model,
                seldon_metrics,
                INPUT_TRANSFORM_METRIC_METHOD_TAG,
                client_response.metrics,
            )

            return construct_response(
                user_model,
                False,
                request,
                client_response.data,
                meta,
                metrics,
                client_response.tags,
            )
        else:
            (features, meta, datadef, data_type) = extract_request_parts_json(request)
            class_names = datadef["names"] if datadef and "names" in datadef else []

            client_response = client_transform_input(
                user_model, features, class_names, meta=meta
            )

            metrics = client_custom_metrics(
                user_model,
                seldon_metrics,
                INPUT_TRANSFORM_METRIC_METHOD_TAG,
                client_response.metrics,
            )

            return construct_response_json(
                user_model,
                False,
                request,
                client_response.data,
                meta,
                metrics,
                client_response.tags,
            )


def transform_output(
    user_model: Any,
    request: Union[prediction_pb2.SeldonMessage, List, Dict],
    seldon_metrics: SeldonMetrics,
) -> Union[prediction_pb2.SeldonMessage, List, Dict]:
    """

    Parameters
    ----------
    user_model
       User defined class to handle transform input
    request
       The incoming request

    Returns
    -------
       The transformed request

    """
    is_proto = isinstance(request, prediction_pb2.SeldonMessage)

    if hasattr(user_model, "transform_output_rest"):
        logger.warning(
            "transform_input_rest is deprecated. Please use transform_input_raw"
        )
        return user_model.transform_output_rest(request)
    elif hasattr(user_model, "transform_output_grpc"):
        logger.warning(
            "transform_input_grpc is deprecated. Please use transform_input_raw"
        )
        return user_model.transform_output_grpc(request)
    else:
        if hasattr(user_model, "transform_output_raw"):
            try:
                response = user_model.transform_output_raw(request)
                handle_raw_custom_metrics(
                    response,
                    seldon_metrics,
                    is_proto,
                    OUTPUT_TRANSFORM_METRIC_METHOD_TAG,
                )
                return response
            except SeldonNotImplementedError:
                pass

        if is_proto:
            (features, meta, datadef, data_type) = extract_request_parts(request)

            client_response = client_transform_output(
                user_model, features, datadef.names, meta=meta
            )

            metrics = client_custom_metrics(
                user_model,
                seldon_metrics,
                OUTPUT_TRANSFORM_METRIC_METHOD_TAG,
                client_response.metrics,
            )

            return construct_response(
                user_model,
                False,
                request,
                client_response.data,
                meta,
                metrics,
                client_response.tags,
            )
        else:
            (features, meta, datadef, data_type) = extract_request_parts_json(request)
            class_names = datadef["names"] if datadef and "names" in datadef else []

            client_response = client_transform_output(
                user_model, features, class_names, meta=meta
            )

            metrics = client_custom_metrics(
                user_model,
                seldon_metrics,
                OUTPUT_TRANSFORM_METRIC_METHOD_TAG,
                client_response.metrics,
            )

            return construct_response_json(
                user_model,
                False,
                request,
                client_response.data,
                meta,
                metrics,
                client_response.tags,
            )


def route(
    user_model: Any,
    request: Union[prediction_pb2.SeldonMessage, List, Dict],
    seldon_metrics: SeldonMetrics,
) -> Union[prediction_pb2.SeldonMessage, List, Dict]:
    """
    Parameters
    ----------
    user_model
       A Seldon user model
    request
       A SelodonMessage proto
    seldon_metrics
        A SeldonMetrics instance
    Returns
    -------
    """
    is_proto = isinstance(request, prediction_pb2.SeldonMessage)

    if hasattr(user_model, "route_rest"):
        logger.warning("route_rest is deprecated. Please use route_raw")
        return user_model.route_rest(request)
    elif hasattr(user_model, "route_grpc"):
        logger.warning("route_grpc is deprecated. Please use route_raw")
        return user_model.route_grpc(request)
    else:
        if hasattr(user_model, "route_raw"):
            try:
                response = user_model.route_raw(request)
                handle_raw_custom_metrics(
                    response, seldon_metrics, is_proto, ROUTER_METRIC_METHOD_TAG
                )
                return response
            except SeldonNotImplementedError:
                pass

        if is_proto:
            (features, meta, datadef, data_type) = extract_request_parts(request)
            client_response = client_route(
                user_model, features, datadef.names, meta=meta
            )
            if not isinstance(client_response.data, int):
                raise SeldonMicroserviceException(
                    "Routing response must be int but got " + str(client_response.data)
                )
            client_response_arr = np.array([[client_response.data]])

            metrics = client_custom_metrics(
                user_model,
                seldon_metrics,
                ROUTER_METRIC_METHOD_TAG,
                client_response.metrics,
            )

            return construct_response(
                user_model,
                False,
                request,
                client_response_arr,
                None,
                metrics,
                client_response.tags,
            )
        else:
            (features, meta, datadef, data_type) = extract_request_parts_json(request)
            class_names = datadef["names"] if datadef and "names" in datadef else []
            client_response = client_route(user_model, features, class_names, meta=meta)
            if not isinstance(client_response.data, int):
                raise SeldonMicroserviceException(
                    "Routing response must be int but got " + str(client_response.data)
                )
            client_response_arr = np.array([[client_response.data]])

            metrics = client_custom_metrics(
                user_model,
                seldon_metrics,
                ROUTER_METRIC_METHOD_TAG,
                client_response.metrics,
            )

            return construct_response_json(
                user_model,
                False,
                request,
                client_response_arr,
                None,
                metrics,
                client_response.tags,
            )


def aggregate(
    user_model: Any,
    request: Union[prediction_pb2.SeldonMessageList, List, Dict],
    seldon_metrics: SeldonMetrics,
) -> Union[prediction_pb2.SeldonMessage, List, Dict]:
    """
    Aggregate a list of payloads

    Parameters
    ----------
    user_model
       A Seldon user model
    request
       SeldonMessage proto
    seldon_metrics
        A SeldonMetrics instance

    Returns
    -------
       Aggregated SeldonMessage proto

    """

    def merge_meta(meta_list):
        tags = {}
        requestPath = {}
        for meta in meta_list:
            if meta:
                tags.update(meta.get("tags", {}))
                requestPath.update(meta.get("requestPath", {}))
        return {"tags": tags, "requestPath": requestPath}

    def merge_metrics(meta_list, custom_metrics):
        metrics = []
        for meta in meta_list:
            if meta:
                metrics.extend(meta.get("metrics", []))
        metrics.extend(custom_metrics)
        return metrics

    is_proto = isinstance(request, prediction_pb2.SeldonMessageList)

    if hasattr(user_model, "aggregate_rest"):
        logger.warning("aggregate_rest is deprecated. Please use aggregate_raw")
        return user_model.aggregate_rest(request)
    elif hasattr(user_model, "aggregate_grpc"):
        logger.warning("aggregate_grpc is deprecated. Please use aggregate_raw")
        return user_model.aggregate_grpc(request)
    else:
        if hasattr(user_model, "aggregate_raw"):
            try:
                response = user_model.aggregate_raw(request)
                handle_raw_custom_metrics(
                    response, seldon_metrics, is_proto, AGGREGATE_METRIC_METHOD_TAG
                )
                return response
            except SeldonNotImplementedError:
                pass

        if is_proto:
            features_list = []
            names_list = []
            meta_list = []

            for msg in request.seldonMessages:
                (features, meta, datadef, data_type) = extract_request_parts(msg)
                features_list.append(features)
                names_list.append(datadef.names)
                meta_list.append(meta)

            client_response = client_aggregate(user_model, features_list, names_list)

            metrics = client_custom_metrics(
                user_model,
                seldon_metrics,
                AGGREGATE_METRIC_METHOD_TAG,
                client_response.metrics,
            )

            return construct_response(
                user_model,
                False,
                request.seldonMessages[0],
                client_response.data,
                merge_meta(meta_list),
                merge_metrics(meta_list, metrics),
                client_response.tags,
            )
        else:
            features_list = []
            names_list = []

            if isinstance(request, list):
                msgs = request
            elif "seldonMessages" in request and isinstance(
                request["seldonMessages"], list
            ):
                msgs = request["seldonMessages"]
            else:
                raise SeldonMicroserviceException(
                    f"Invalid request data type: {request}"
                )

            meta_list = []
            for msg in msgs:
                (features, meta, datadef, data_type) = extract_request_parts_json(msg)
                class_names = datadef["names"] if datadef and "names" in datadef else []
                features_list.append(features)
                names_list.append(class_names)
                meta_list.append(meta)

            client_response = client_aggregate(user_model, features_list, names_list)

            metrics = client_custom_metrics(
                user_model,
                seldon_metrics,
                AGGREGATE_METRIC_METHOD_TAG,
                client_response.metrics,
            )

            return construct_response_json(
                user_model,
                False,
                msgs[0],
                client_response.data,
                merge_meta(meta_list),
                merge_metrics(meta_list, metrics),
                client_response.tags,
            )


def health_status(
    user_model: Any, seldon_metrics: SeldonMetrics
) -> Union[prediction_pb2.SeldonMessage, List, Dict]:
    """
    Call the user model to check the health of the model

    Parameters
    ----------
    user_model
       User defined class instance
    seldon_metrics
        A SeldonMetrics instance

    Returns
    -------
      Health check output
    """

    if hasattr(user_model, "health_status_raw"):
        try:
            return user_model.health_status_raw()
        except SeldonNotImplementedError:
            pass

    client_response = client_health_status(user_model)
    metrics = client_custom_metrics(
        user_model, seldon_metrics, HEALTH_METRIC_METHOD_TAG
    )

    return construct_response_json(
        user_model, False, {}, client_response, None, metrics
    )


def init_metadata(user_model: Any) -> Dict:
    """
    Call the user model to get the model init_metadata

    Parameters
    ----------
    user_model
        User defined class instance

    Returns
    -------
        Validated model metadata
    """
    # meta_user: load metadata defined in the user_model instance
    if hasattr(user_model, "init_metadata"):
        try:
            meta_user = user_model.init_metadata()
        except SeldonNotImplementedError:
            meta_user = {}
            pass
    else:
        meta_user = {}

    if not isinstance(meta_user, dict):
        logger.error("init_metadata must return dict")
        meta_user = {}

    # meta_env: load metadata from environmental variable
    try:
        meta_env = yaml.safe_load(os.environ.get("MODEL_METADATA", "{}"))
    except yaml.YAMLError as e:
        logger.error(f"Reading metadata from MODEL_METADATA env variable failed: {e}")
        meta_env = {}

    meta = {**meta_user, **meta_env}

    try:
        return validate_model_metadata(meta)
    except SeldonInvalidMetadataError as e:
        logger.error(f"Metadata validation error\n{e}")
        logger.error(f"Failed to validate metadata {meta}")
        return None



from seldon_core.imports_helper import _TF_PRESENT

if _TF_PRESENT:
    # Let tensorflow shadow these imports if present
    from tensorflow.core.framework import tensor_pb2, types_pb2, tensor_shape_pb2



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from seldon_core.proto import prediction_pb2 as proto_dot_prediction__pb2

GRPC_GENERATED_VERSION = '1.70.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in proto/prediction_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class GenericStub(object):
    """[END Messages]

    [START Services]

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TransformInput = channel.unary_unary(
                '/seldon.protos.Generic/TransformInput',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.TransformOutput = channel.unary_unary(
                '/seldon.protos.Generic/TransformOutput',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.Route = channel.unary_unary(
                '/seldon.protos.Generic/Route',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.Aggregate = channel.unary_unary(
                '/seldon.protos.Generic/Aggregate',
                request_serializer=proto_dot_prediction__pb2.SeldonMessageList.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.SendFeedback = channel.unary_unary(
                '/seldon.protos.Generic/SendFeedback',
                request_serializer=proto_dot_prediction__pb2.Feedback.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)


class GenericServicer(object):
    """[END Messages]

    [START Services]

    """

    def TransformInput(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TransformOutput(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Route(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Aggregate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendFeedback(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GenericServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TransformInput': grpc.unary_unary_rpc_method_handler(
                    servicer.TransformInput,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'TransformOutput': grpc.unary_unary_rpc_method_handler(
                    servicer.TransformOutput,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'Route': grpc.unary_unary_rpc_method_handler(
                    servicer.Route,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'Aggregate': grpc.unary_unary_rpc_method_handler(
                    servicer.Aggregate,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessageList.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'SendFeedback': grpc.unary_unary_rpc_method_handler(
                    servicer.SendFeedback,
                    request_deserializer=proto_dot_prediction__pb2.Feedback.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Generic', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Generic', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Generic(object):
    """[END Messages]

    [START Services]

    """

    @staticmethod
    def TransformInput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Generic/TransformInput',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TransformOutput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Generic/TransformOutput',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Route(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Generic/Route',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Aggregate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Generic/Aggregate',
            proto_dot_prediction__pb2.SeldonMessageList.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendFeedback(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Generic/SendFeedback',
            proto_dot_prediction__pb2.Feedback.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class ModelStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Predict = channel.unary_unary(
                '/seldon.protos.Model/Predict',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.SendFeedback = channel.unary_unary(
                '/seldon.protos.Model/SendFeedback',
                request_serializer=proto_dot_prediction__pb2.Feedback.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.Metadata = channel.unary_unary(
                '/seldon.protos.Model/Metadata',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonModelMetadata.FromString,
                _registered_method=True)


class ModelServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Predict(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendFeedback(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Metadata(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ModelServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Predict': grpc.unary_unary_rpc_method_handler(
                    servicer.Predict,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'SendFeedback': grpc.unary_unary_rpc_method_handler(
                    servicer.SendFeedback,
                    request_deserializer=proto_dot_prediction__pb2.Feedback.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'Metadata': grpc.unary_unary_rpc_method_handler(
                    servicer.Metadata,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonModelMetadata.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Model', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Model', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Model(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Predict(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Model/Predict',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendFeedback(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Model/SendFeedback',
            proto_dot_prediction__pb2.Feedback.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Metadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Model/Metadata',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            proto_dot_prediction__pb2.SeldonModelMetadata.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class RouterStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Route = channel.unary_unary(
                '/seldon.protos.Router/Route',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.SendFeedback = channel.unary_unary(
                '/seldon.protos.Router/SendFeedback',
                request_serializer=proto_dot_prediction__pb2.Feedback.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)


class RouterServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Route(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendFeedback(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RouterServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Route': grpc.unary_unary_rpc_method_handler(
                    servicer.Route,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'SendFeedback': grpc.unary_unary_rpc_method_handler(
                    servicer.SendFeedback,
                    request_deserializer=proto_dot_prediction__pb2.Feedback.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Router', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Router', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Router(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Route(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Router/Route',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendFeedback(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Router/SendFeedback',
            proto_dot_prediction__pb2.Feedback.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class TransformerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TransformInput = channel.unary_unary(
                '/seldon.protos.Transformer/TransformInput',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)


class TransformerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def TransformInput(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TransformerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TransformInput': grpc.unary_unary_rpc_method_handler(
                    servicer.TransformInput,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Transformer', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Transformer', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Transformer(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def TransformInput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Transformer/TransformInput',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class OutputTransformerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TransformOutput = channel.unary_unary(
                '/seldon.protos.OutputTransformer/TransformOutput',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)


class OutputTransformerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def TransformOutput(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_OutputTransformerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TransformOutput': grpc.unary_unary_rpc_method_handler(
                    servicer.TransformOutput,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.OutputTransformer', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.OutputTransformer', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class OutputTransformer(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def TransformOutput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.OutputTransformer/TransformOutput',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class CombinerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Aggregate = channel.unary_unary(
                '/seldon.protos.Combiner/Aggregate',
                request_serializer=proto_dot_prediction__pb2.SeldonMessageList.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)


class CombinerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Aggregate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_CombinerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Aggregate': grpc.unary_unary_rpc_method_handler(
                    servicer.Aggregate,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessageList.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Combiner', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Combiner', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Combiner(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Aggregate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Combiner/Aggregate',
            proto_dot_prediction__pb2.SeldonMessageList.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class SeldonStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Predict = channel.unary_unary(
                '/seldon.protos.Seldon/Predict',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.SendFeedback = channel.unary_unary(
                '/seldon.protos.Seldon/SendFeedback',
                request_serializer=proto_dot_prediction__pb2.Feedback.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.ModelMetadata = channel.unary_unary(
                '/seldon.protos.Seldon/ModelMetadata',
                request_serializer=proto_dot_prediction__pb2.SeldonModelMetadataRequest.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonModelMetadata.FromString,
                _registered_method=True)
        self.GraphMetadata = channel.unary_unary(
                '/seldon.protos.Seldon/GraphMetadata',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonGraphMetadata.FromString,
                _registered_method=True)


class SeldonServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Predict(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendFeedback(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ModelMetadata(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GraphMetadata(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SeldonServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Predict': grpc.unary_unary_rpc_method_handler(
                    servicer.Predict,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'SendFeedback': grpc.unary_unary_rpc_method_handler(
                    servicer.SendFeedback,
                    request_deserializer=proto_dot_prediction__pb2.Feedback.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'ModelMetadata': grpc.unary_unary_rpc_method_handler(
                    servicer.ModelMetadata,
                    request_deserializer=proto_dot_prediction__pb2.SeldonModelMetadataRequest.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonModelMetadata.SerializeToString,
            ),
            'GraphMetadata': grpc.unary_unary_rpc_method_handler(
                    servicer.GraphMetadata,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonGraphMetadata.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Seldon', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Seldon', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Seldon(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Predict(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Seldon/Predict',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendFeedback(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Seldon/SendFeedback',
            proto_dot_prediction__pb2.Feedback.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ModelMetadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Seldon/ModelMetadata',
            proto_dot_prediction__pb2.SeldonModelMetadataRequest.SerializeToString,
            proto_dot_prediction__pb2.SeldonModelMetadata.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GraphMetadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Seldon/GraphMetadata',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            proto_dot_prediction__pb2.SeldonGraphMetadata.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)
